/*
 * ArpTelnet Copyright (c)1997-98 by Dianne Hackborn.
 * All rights reserved.
 *
 * Based (increasingly loosly) on WebTerm, a Web-based terminal
 * applet written in Java, which is
 * Copyright (C)1996 by the National Alliance for Computational
 * Science and Engineering (NACSE).
 * See <URL:http://www.nacse.org/> for more information.
 *
 * This code is not public domain, nor freely distributable.
 * Please direct any questions or requests to Dianne Hackborn,
 * at <hackbod@lucent.com> or <hackbod@angryredplanet.com>.
 *
 */

#ifndef _APPLICATION_H
#include <be/app/Application.h>
#endif

#ifndef _SCREEN_H
#include <be/interface/Screen.h>
#endif

#ifndef _RECT_H
#include <be/interface/Rect.h>
#endif

#ifndef _ALERT_H
#include <be/interface/Alert.h>
#endif

#ifndef _BITMAP_H
#include <be/interface/Bitmap.h>
#endif

#ifndef _FILE_PANEL_H
#include <be/storage/FilePanel.h>
#endif

#ifndef _MIME_H
#include <be/storage/Mime.h>
#endif

#ifndef _APPFILEINFO_H
#include <be/storage/AppFileInfo.h>
#endif

#ifndef _FILE_H
#include <be/storage/File.h>
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include <ArpKernel/ArpDebug.h>
#endif

#ifndef ARPKERNEL_ARPMULTIDIR_H
#include <ArpKernel/ArpMultiDir.h>
#endif

#ifndef ARPNETWORK_ARPTELNET_H
#include <ArpNetwork/ArpTelnet.h>
#endif

#ifndef ARPKERNEL_ARPSTRING_H
#include <ArpKernel/ArpString.h>
#endif

#ifndef ARPKERNEL_ARPABOUTWINDOW_H
#include <ArpKernel/ArpAboutWindow.h>
#endif

#ifndef MAIN_H
#include "Main.h"
#endif

#ifndef TERMINALWIN_H
#include "TerminalWin.h"
#endif

ArpMOD();

// avoid a compiler bug...
#undef ArpDEBUG

#include <stdlib.h>
#include <getopt.h>

const char* ArpAppSig = "application/x-vnd.ARP-ArpTelnet";
const char* ArpTelnetConfigMime = "application/x-vnd.ARP.ArpTelnetConfig";
const char* ArpTelnetURLMime = "application/x-vnd.Be.URL.telnet";

#if defined(ArpDEBUG)
#include <ArpCollections/ArpSTLVector.h>

static void ShowURLParts(const ArpString& url)
{
	ArpString proto, addr, path, loc;
	ArpSTLVector<ArpString> params;
	
	ArpD(cdb << ADH << "--- Parsing URL: " << url << endl);
	status_t err = ArpParseURL(url, &proto, &addr, &path, &params, &loc);
	if( err ) {
		ArpD(cdb << ADH << "Error during parse!  Code: " << err << endl);
	}
	if( !proto.IsNull() ) {
		ArpD(cdb << ADH << "Protocol: " << proto << endl);
	} else {
		 ArpD(cdb << ADH << "No protocol specified." << endl);
	}
	if( !addr.IsNull() ) {
		ArpD(cdb << ADH << "Address: " << addr << endl);
	} else {
		ArpD(cdb << ADH << "No address specified." << endl);
	}
	if( !path.IsNull() ) {
		ArpD(cdb << ADH << "Path: " << path << endl);
	} else {
		ArpD(cdb << ADH << "No path specified." << endl);
	}
	
	if( params.size() > 0 ) {
		for( size_t i = 0; i<params.size(); i++ ) {
			ArpD(cdb << ADH << "Parameter #" << i << ": " << params[i] << endl);
		}
	} else {
		ArpD(cdb << ADH << "No parameters specified." << endl);
	}
	
	if( !loc.IsNull() ) {
		ArpD(cdb << ADH << "Location: " << loc << endl);
	} else {
		ArpD(cdb << ADH << "No location specified." << endl);
	}
}
#endif

struct teststr1 {
	uint8 space1;
	uint32 bits32;
	uint8 space2;
	uint64 bits64;
	uint8 space3;
	float real32;
	uint8 space4;
	double real64;
};

struct teststr2 {
	uint64 bits64;
	double real64;
	uint32 bits32;
	float real32;
	uint8 space1;
	uint8 space2;
	uint8 space3;
	uint8 space4;
};

main()
{	
	printf("Size1 = %ld, size2 = %ld\n",
			sizeof(teststr1), sizeof(teststr2));
	
	MainApp *myApplication;

	myApplication = new MainApp();

#if defined(ArpDEBUG)
	// Here is an example using ArpMultiDir...  originally written
	// just for my own testing, but I thought I'd leave it in since
	// the class is otherwise rather undocumented. ;)
	BPath path;
	if( find_directory(B_BEOS_ADDONS_DIRECTORY,&path) == B_OK ) {
		printf("BEOS ADDONS: %s\n",path.Path());
	}
	if( find_directory(B_COMMON_ADDONS_DIRECTORY,&path) == B_OK ) {
		printf("COMMON ADDONS: %s\n",path.Path());
	}
	if( find_directory(B_USER_ADDONS_DIRECTORY,&path) == B_OK ) {
		printf("USER ADDONS: %s\n",path.Path());
	}
	
	ArpMultiDir mdir;
	mdir.AddDirectory(B_BEOS_ADDONS_DIRECTORY,"ARP/TermEmulator");
	mdir.AddDirectory(B_COMMON_ADDONS_DIRECTORY,"ARP/TermEmulator");
	mdir.AddDirectory(B_USER_ADDONS_DIRECTORY,"ARP/TermEmulator");
	mdir.AddEnvVar("ADDON_PATH","ARP/TermEmulator");
	BEntry entry;
	mdir.Rewind();
	while( mdir.GetNextEntry(&entry,true) == B_NO_ERROR ) {
		BPath path;
		entry.GetPath(&path);
		printf("In dir %s found %s\n","<unknown>",path.Path());
	}
	
	ShowURLParts("telnet:");
	ShowURLParts("telnet://adam.enteract.com/");
	ShowURLParts("telnet://adam.enteract.com:23/");
	ShowURLParts("telnet://adam.enteract.com:23/%3Bhackbod");
	ShowURLParts("www.enteract.com");
	ShowURLParts("http://www.enteract.com");
	ShowURLParts("http://www.enteract.com/");
	ShowURLParts("http:/root/dir");
	ShowURLParts("http:relative/dir");
	ShowURLParts("http:relative/dir?with=param?more=param2");
	ShowURLParts("/root/dir");
	ShowURLParts("relative/dir");
	ShowURLParts("relative/dir?with=param?more=param2");
	ShowURLParts("http://www.enteract.com/full/url?with=param?more=param2");
	ShowURLParts("#piece");
	ShowURLParts("http://www.enteract.com/full/url?with=param?more=param2#piece");
#endif

	myApplication->Run();
	
	delete(myApplication);
	return(0);
}

MainApp::MainApp()
	: BApplication(ArpAppSig),
	  cmdline(false), mMadeWindow(false),
	  mOpenPanel(0), mSavePanel(0)
{
}

MainApp::~MainApp()
{
	delete mOpenPanel;
	delete mSavePanel;
}

bool MainApp::QuitRequested(void)
{
	if( mAboutWin.IsValid() ) mAboutWin.SendMessage(B_QUIT_REQUESTED);
	return true;
}

static void InstallDescription(BMimeType& type, const char* sdescr, const char* ldescr)
{
	char cur[B_MIME_TYPE_LENGTH];
	if( sdescr ) {
		cur[0] = 0;
		ArpD(cdb << ADH << "Checking short description...\n");
		status_t err = type.GetShortDescription(cur);
		if( err != B_OK || cur[0] == 0 ) {
			type.SetShortDescription(sdescr);
			ArpD(cdb << ADH << "Installing mine.\n");
		}
		ArpD(cdb << ADH << "Req short description: err=" << err
				<< ", string=" << &cur[0] << endl);
	}
	if( ldescr ) {
		cur[0] = 0;
		ArpD(cdb << ADH << "Checking long description...\n");
		status_t err = type.GetLongDescription(cur);
		if( err != B_OK || cur[0] == 0 ) {
			type.SetLongDescription(ldescr);
			ArpD(cdb << ADH << "Installing mine.\n");
		}
		ArpD(cdb << ADH << "Req long description: err=" << err
				<< ", string=" << &cur[0] << endl);
	}
}

static void InstallPrefApp(BMimeType& type, const char* app)
{
	char cur[B_MIME_TYPE_LENGTH];
	cur[0] = 0;
	ArpD(cdb << ADH << "Checking preferred app...\n");
	status_t err = type.GetPreferredApp(cur);
	if( err != B_OK || cur[0] == 0 ) {
		type.SetPreferredApp(app);
		ArpD(cdb << ADH << "Installing mine.\n");
	}
	ArpD(cdb << ADH << "Req preferred app: err=" << err
			<< ", string=" << &cur[0] << endl);
}

static void InstallAppHint(BMimeType& type, const entry_ref* ref)
{
	entry_ref cur;
	ArpD(cdb << ADH << "Checking application hint...\n");
	status_t err = type.GetAppHint(&cur);
	if( err != B_OK || cur == entry_ref() ) {
		type.SetAppHint(ref);
		ArpD(cdb << ADH << "Installing mine.\n");
	}
	ArpD(cdb << ADH << "Req app hint: err=" << err
			<< ", leaf=" << (cur.name ? cur.name : "<NULL>") << endl);
}

static void InstallAppIcon(BMimeType& type, const char* file_type,
							icon_size size, const entry_ref* app)
{
	BBitmap* cur = new BBitmap(BRect(0,0,size-1,size-1), B_CMAP8);
	ArpD(cdb << ADH << "Checking " << file_type << " "
			<< int(size) << "x" << int(size) << " bitmap..." << endl);
	status_t err = type.GetIconForType(file_type, cur, size);
	if( err != B_OK ) {
		ArpD(cdb << ADH << "Need to install my icon.\n");
		BFile file(app, B_READ_ONLY);
		BAppFileInfo afi(&file);
		if( (err=afi.InitCheck()) == B_OK ) {
			BBitmap* icon = new BBitmap(BRect(0,0,size-1,size-1), B_CMAP8);
			if( (err=afi.GetIcon(icon, size)) == B_OK ) {
				err = type.SetIconForType(file_type, icon, size);
				ArpD(cdb << ADH << "Installed.\n");
			}
			delete icon;
		}
	}
	ArpD(cdb << ADH << "Req small bitmap: err=" << err << endl);
	delete cur;
}
							
static void InstallAppIcon(BMimeType& type, const char* file_type,
							const entry_ref* app)
{
	InstallAppIcon(type, file_type, B_MINI_ICON, app);
	InstallAppIcon(type, file_type, B_LARGE_ICON, app);
}

struct mimeattr_info {
	const char* name;
	const char* public_name;
	type_code type;
	int32 width;
	int32 alignment;
	bool viewable;
	bool editable;
	bool extra;
};

static void InstallAttributes(BMimeType& type, mimeattr_info* attrs)
{
	ArpD(cdb << ADH << "Installing attributes into " << type.Type() << endl);
	
	BMessage msg;
	type.GetAttrInfo(&msg);
	
	ArpD(cdb << ADH << "Current attributes: " << msg << endl);
	
	bool changed = false;
	
	while( attrs && attrs->name ) {
		type_code type;
		int32 c;
		bool found = false;
		
		ArpD(cdb << ADH << "Checking attr " << attrs->name << endl);
		
		// See if this attribute already exists.
		if( msg.GetInfo("attr:name", &type, &c) == B_OK ) {
			ArpD(cdb << ADH << "Found " << c << " existing attrs." << endl);
			for( int32 i=0; !found && i<c; i++ ) {
				const char* name;
				if( msg.FindString("attr:name", i, &name) == B_OK ) {
					ArpD(cdb << ADH << "Comparing #" << i << ": "
									<< attrs->name << " with " << name << endl);
					if( strcmp(name, attrs->name) == 0 ) found = true;
				}
			}
		}
		
		// Didn't find it?
		if( !found ) {
			ArpD(cdb << ADH << "Adding this attribute." << endl);
			if( msg.AddString("attr:name", attrs->name) ) return;
			if( msg.AddString("attr:public_name", attrs->public_name) ) return;
			if( msg.AddInt32("attr:type", attrs->type) ) return;
			if( msg.AddInt32("attr:width", attrs->width) ) return;
			if( msg.AddInt32("attr:alignment", attrs->alignment) ) return;
			if( msg.AddBool("attr:viewable", attrs->viewable) ) return;
			if( msg.AddBool("attr:editable", attrs->editable) ) return;
			if( msg.AddBool("attr:extra", attrs->extra) ) return;
			changed = true;
		}
		
		attrs++;
	}
	
	if( changed ) {
		ArpD(cdb << ADH << "Installing new attributes: " << msg << endl);
		type.SetAttrInfo(&msg);
	} else {
		ArpD(cdb << ADH << "No new attributes!" << endl);
	}
}

static mimeattr_info telnet_attrs[] = {
	{	"META:url",			"URL",
		B_STRING_TYPE,		170,			B_ALIGN_LEFT,
		true,				true,			false
	},
	{	"META:keyw",		"Keywords",
		B_STRING_TYPE,		130,			B_ALIGN_LEFT,
		true,				true,			false
	},
	{	"META:title",		"Title",
		B_STRING_TYPE,		130,			B_ALIGN_LEFT,
		true,				true,			false
	},
	{ 0 }
};

void MainApp::AddMimeTypes()
{
	BMimeType mt;
	
	mt.SetTo(ArpTelnetConfigMime);
	ArpD(cdb << ADH << "Setting up config MIME type...\n");
	if( mt.InitCheck() == B_OK ) {
		if( !mt.IsInstalled() ) {
			ArpD(cdb << ADH << "Installing config type.\n");
			ArpD(cdb << ADH << "Config type not installed, installing it." << endl);
			mt.Install();
		}
		InstallDescription(mt, "ArpTelnet Configuration",
						   "BMessage archive of ArpTelnet program settings.");
		InstallPrefApp(mt, ArpAppSig);
	}
	mt.Unset();
	
	mt.SetTo(ArpTelnetURLMime);
	ArpD(cdb << ADH << "Setting up telnet URL MIME type...\n");
	if( mt.InitCheck() == B_OK ) {
		ArpD(cdb << ADH << "Checking URL telnet type..." << endl);
		if( !mt.IsInstalled() ) {
			ArpD(cdb << ADH << "Installing telnet URL type.\n");
			ArpD(cdb << ADH << "Telnet type not installed, installing it." << endl);
			mt.Install();
		}
		InstallDescription(mt,"URL: telnet",
						   "Bookmark of a URL for the telnet protocol.");
		InstallPrefApp(mt, ArpAppSig);
		InstallAttributes(mt, telnet_attrs);
		
		ArpD(cdb << ADH << "Checking preferred app." << endl);
		char prefapp[B_MIME_TYPE_LENGTH];
		prefapp[0] = 0;
		if( mt.GetPreferredApp(prefapp) == B_OK ) {
			ArpD(cdb << ADH << "Current preferred app is: " << prefapp << endl);
			if( strcmp(prefapp, ArpAppSig) != 0 ) {
				ArpD(cdb << ADH << "Asking user to replace it." << endl);
				BAlert* myAlert = new BAlert("Make Preferred App",
					"ArpTelnet is not currently registered as the preferred "
					"application for 'telnet' URLs.  Make it the new preferred "
					"application?",
					"No", "Don't Ask", "Yes",
					B_WIDTH_FROM_WIDEST, B_WARNING_ALERT);
				myAlert->SetShortcut(0, B_ESCAPE);
				int cmd = myAlert->Go();
				if( cmd == 2 ) {
					mt.SetPreferredApp(ArpAppSig);
				}
			}
		}
	}
	mt.Unset();
	
	mt.SetTo(ArpAppSig);
	ArpD(cdb << ADH << "Updating application MIME type...\n");
	if( mt.InitCheck() == B_OK ) {
		if( !mt.IsInstalled() ) {
			// This shouldn't happen, I don't think...
			ArpD(cdb << ADH << "Installing application type.\n");
			ArpD(cdb << ADH << "Application type not installed, installing it." << endl);
			mt.Install();
		}
		InstallDescription(mt, "ArpTelnet", 0);
		InstallPrefApp(mt, ArpAppSig);
		app_info ai;
		if( be_app->GetAppInfo(&ai) == B_OK ) {
			InstallAppHint(mt, &ai.ref);
			InstallAppIcon(mt, ArpTelnetConfigMime, &ai.ref);
			InstallAppIcon(mt, ArpTelnetURLMime, &ai.ref);
		}
	}
	mt.Unset();
	
}

bool MainApp::MakeTermWindow(const char* url, const entry_ref& settings)
{
	AddMimeTypes();
	
	TerminalWin* termwin;
	ArpD(cdb << ADH << "Creating terminal window..." << endl);
	termwin = new TerminalWin(BPoint(20,20),
								"ArpTelnet v" PROGRAM_VERSION,
								entry_ref(),
								"About ArpTelnet" B_UTF8_ELLIPSIS);
	
	// make window visible
	if(termwin) {
		termmsg = BMessenger(termwin);
		
		BScreen screen(termwin);
		if( screen.IsValid() ) {
			BRect wfrm = termwin->Frame();
			BRect sfrm = screen.Frame();
			float height = wfrm.Height();
			float width = wfrm.Width();
			if( height > sfrm.Height() ) height = sfrm.Height();
			if( width > sfrm.Width() ) width = sfrm.Width();
			if( !termwin->HasPosition() ) {
				wfrm.left = sfrm.left + (sfrm.Width()-width)/2;
				wfrm.top = sfrm.top + (sfrm.Height()-height)/2;
			}
			if( wfrm.left < (-width+10) ) wfrm.left = -width+10;
			else if( wfrm.left > (sfrm.Width()-10) ) {
				wfrm.left = sfrm.Width() - 10;
			}
			if( wfrm.top < (-height+10) ) wfrm.top = -height+10;
			else if( wfrm.top > (sfrm.Height()-10) ) {
				wfrm.top = sfrm.Height() - 10;
			}
			termwin->MoveTo( wfrm.left, wfrm.top );
			termwin->ResizeTo(width,height);
		}
		
		ArpD(cdb << ADH << "Reading configuration for terminal win..." << endl);
		termwin->OpenConfiguration(settings, url ? false : true);
		
		ArpD(cdb << ADH << "Creating remote to: " << (url ? url : "<null>")
						<< endl);
		if( url ) termwin->SetRemote(url);
		
		ArpD(cdb << ADH << "Showing terminal window..." << endl);
		termwin->Show();
		return true;
	}
	
	return false;
}

void MainApp::AboutRequested(void)
{
	if( !mAboutWin.IsValid() ) {
		ArpAboutWindow* win = 
			new ArpAboutWindow(0, "ArpTelnet", 0, __DATE__,
					"ArpTelnet version " PROGRAM_VERSION
					" / " __DATE__ "\n"
					B_UTF8_COPYRIGHT "1999 Angry Red Planet Software.\n\n"
					
					"A freeware terminal app featuring add-on\n"
					"emulators and devices.  Includes VT200,\n"
					"VT100, VT52, and XTerm emulators;\n"
					"and Telnet and Shell devices.\n\n"
					
					"For more info and the latest version, see\n"
					"<URL:http://www.angryredplanet.com/>.\n\n"
					
					"Written by Dianne Hackborn\n"
					"(email: hackbod@angryredplanet.com)\n\n"
					
					"Based (loosly) on WebTerm, a Java terminal\n"
					"applet, whose copyright is held by the\n"
					"Northwest Alliance for Computational\n"
					"Science and Engineering.  (On the web\n"
					"at <URL:http://www.nacse.org/>.)\n\n"
					
					"Thanks to:\n\n"
					
					"Brian J. Swetland\n"
					"For the basic code to the shell device.\n\n"
					
					"Jon Watte\n"
					"For many hints on using sockets under BeOS.\n\n"
					
					"Gustav Kalvesten\n"
					"For reporting a problem with the XTerm\n"
					"title text sequences.\n\n"

					"Tom Spindler\n"
					"For pointing me to \"vttest\" and sending\n"
					"over a copy.\n\n"

					"Rich $alz and Per Lindberg\n"
					"For vttest, which has made the VT emulation\n"
					"much more solid...  it's not completely there\n"
					"yet, but thanks to them getting much better.\n\n");

		mAboutWin = BMessenger(win);
		win->Show();
	}
}

void MainApp::DispatchMessage(BMessage *message,
								  BHandler *handler)
{
	ArpD(cdb << ADH << "MainApp::DispatchMessage: " <<
				*message << endl);
	inherited::DispatchMessage(message,handler);
}

void MainApp::MessageReceived(BMessage* message)
{
	if( !message ) return;
	
	switch( message->what ) {
		case TERMINALWIN_OPENSETTINGS_MSG: {
			BMessenger target;
			entry_ref curFile;
			BEntry curDir;
			status_t err = message->FindMessenger("who", &target);
			if( message->FindRef("current", &curFile) == B_OK ) {
				curDir.SetTo(&curFile);
				if( curDir.InitCheck() == B_OK ) curDir.GetParent(&curDir);
			}
			if( err ) break;
			
			if( !mOpenPanel ) mOpenPanel = new BFilePanel(B_OPEN_PANEL, 0, 0, 0,
														  false);
			if( mOpenPanel->IsShowing() ) return;
			if( mOpenPanel->Window() ) {
				mOpenPanel->Window()->SetTitle("Open ArpTelnet Session"
											   B_UTF8_ELLIPSIS);
			}
			
			mOpenPanel->SetTarget(target);
			if( curDir.InitCheck() == B_OK ) {
				mOpenPanel->SetPanelDirectory(&curDir);
			}
			mOpenPanel->Show();
		} break;
		case TERMINALWIN_SAVESETTINGS_MSG: {
			BMessenger target;
			entry_ref curFile;
			BEntry curDir;
			status_t err = message->FindMessenger("who", &target);
			if( message->FindRef("current", &curFile) == B_OK ) {
				curDir.SetTo(&curFile);
				if( curDir.InitCheck() == B_OK ) curDir.GetParent(&curDir);
			}
			if( err ) break;
			
			if( !mSavePanel ) mSavePanel = new BFilePanel(B_SAVE_PANEL, 0, 0, 0,
														  false);
			if( mSavePanel->IsShowing() ) return;
			if( mSavePanel->Window() ) {
				mSavePanel->Window()->SetTitle("Save ArpTelnet Session"
											   B_UTF8_ELLIPSIS);
			}
			mSavePanel->SetTarget(target);
			if( curDir.InitCheck() == B_OK ) {
				mSavePanel->SetPanelDirectory(&curDir);
			}
			if( curFile.name ) mSavePanel->SetSaveText(curFile.name);
			mSavePanel->Show();
		} break;
		default:
			inherited::MessageReceived(message);
	}
}

static void Usage(int /*argc*/, char **argv)
{
	printf("Usage: %s [--file | --url | --telnet] <destination>\n"
		   "\tThe <destination> argument depends on the preceeding flag:\n"
		   "\t--file: Path to settings file\n"
		   "\t--url: URL to open, e.g. telnet://myaddress.com/\n"
		   "\t--telnet: Address for telnet connect, with optional port number\n",
	       argv[0]);
}

void MainApp::ArgvReceived(int32 argc, char** argv)
{
	int c;
	enum ArgType { ARG_UNKNOWN, ARG_FILE, ARG_URL, ARG_TELNET };
	ArgType argtype = ARG_UNKNOWN;
	
	ArpParseDBOpts(argc, argv);
	
	// Parse options
	while(1) {
		int option_index = 0;
		static struct option long_options[] = {
			{"file",	no_argument,	0,	'f'},
			{"url",		no_argument,	0,	'u'},
			{"telnet",	no_argument,	0,	't'},
			{"help",	no_argument,	0,	'h'},
			{0,0,0,0}
		};
		c = getopt_long (argc, argv, "futh", long_options, &option_index);
		if (c == -1)
			break;
		switch(c) {
			case 'f':
				argtype = ARG_FILE;
				break;
			case 'u':
				argtype = ARG_URL;
				break;
			case 't':
				argtype = ARG_TELNET;
				break;
			case 'h':
			case '?':
				Usage(argc,argv);
				return;
			default:
				printf ("?? getopt returned character code 0%o ??\n", c);
		}
	}
	
	// Process command line arguments:  Connect to given session.
	if( optind < argc ) {
		// Is this an existing file?
		if( argtype == ARG_FILE ) {
			ArpD(cdb << ADH << "Creating terminal with file "
					<< argv[optind] << endl);
			BEntry entry(argv[optind], true);
			status_t err = entry.InitCheck();
			entry_ref ref;
			if( !err ) err = entry.GetRef(&ref);
			if( err ) {
				printf("Error opening file %s:\n%s\n",
					   argv[optind], strerror(err));
				return;
			}
			ArpD(cdb << ADH << "File entry: device=" << ref.device
						<< ", dir=" << ref.directory
						<< ", name=" << ref.name << endl);
			if( !mMadeWindow ) {
				mMadeWindow = MakeTermWindow(0, ref);
			} else {
				// Send a new settings file.
				BMessage connmsg(ARPTELNET_CONNECT_MSG);
				connmsg.AddRef("refs", &ref);
				termmsg.SendMessage(&connmsg);
			}
			optind++;
		
		// Is this a URL?
		} else if( argtype == ARG_URL ||
					(argtype == ARG_UNKNOWN && strchr(argv[optind],':') != NULL) ) {
			if( !mMadeWindow ) {
				mMadeWindow = MakeTermWindow(argv[optind]);
			} else {
				// Send a raw url.
				BMessage connmsg(ARPTELNET_CONNECT_MSG);
				connmsg.AddString("url", argv[optind]);
				termmsg.SendMessage(&connmsg);
			}
			optind++;
		
		// Otherwise, interpret as telnet address and optional port number.
		} else {
			ArpString url;
			url = ArpLIT("telnet://") + argv[optind];
			optind++;
			if( optind < argc ) {
				url += ArpLIT(":") + argv[optind] + ":";
				optind++;
			}
			if( !mMadeWindow ) {
				mMadeWindow = MakeTermWindow(url);
			} else {
				BMessage connmsg(ARPTELNET_CONNECT_MSG);
				connmsg.AddString("url", url);
				termmsg.SendMessage(&connmsg);
			}
		}
		cmdline = true;
		
	} else if( argtype != ARG_UNKNOWN ) {
		Usage(argc,argv);
	}
}

void MainApp::RefsReceived(BMessage* message)
{
	uint32 type;
	int32 count;
	entry_ref ref;
	
	if( !message ) return;
	
	message->GetInfo("refs", &type, &count); 
	if ( type != B_REF_TYPE ) return; 
	
	if( count > 0 ) {
		if( message->FindRef("refs", 0, &ref) == B_OK ) {
			if( !mMadeWindow ) {
				mMadeWindow = MakeTermWindow(0, ref);
			} else {
				// Send a new settings file.
				BMessage connmsg(ARPTELNET_CONNECT_MSG);
				connmsg.AddRef("refs", &ref);
				termmsg.SendMessage(&connmsg);
			}
		}
	}
}

void MainApp::ReadyToRun()
{
	if( !mMadeWindow ) mMadeWindow = MakeTermWindow(0);
}
