/*
	
	SessionSettings.cpp
	
	Copyright (c)1999 by Angry Red Planet.

	This code is distributed under a modified form of the
	Artistic License.  A copy of this license should have
	been included with it; if this wasn't the case, the
	entire package can be obtained at
	<URL:http://www.angryredplanet.com/>.
*/

#ifndef _BUTTON_H
#include <be/interface/Button.h>
#endif

#ifndef _MENUFIELD_H
#include <be/interface/MenuField.h>
#endif

#ifndef _POPUPMENU_H
#include <be/interface/PopUpMenu.h>
#endif

#ifndef _MENUITEM_H
#include <be/interface/MenuItem.h>
#endif

#ifndef _COLORCONTROL_H
#include <be/interface/ColorControl.h>
#endif

#ifndef _STRINGVIEW_H
#include <be/interface/StringView.h>
#endif

#ifndef _CHECKBOX_H
#include <be/interface/CheckBox.h>
#endif

#ifndef _SCREEN_H
#include <be/interface/Screen.h>
#endif

#ifndef _AUTOLOCK_H
#include <be/support/Autolock.h>
#endif

#ifndef _UTF8_H
#include <be/support/UTF8.h>
#endif

#ifndef ARPKERNEL_ARPCOLOR_H
#include <ArpKernel/ArpColor.h>
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include <ArpKernel/ArpDebug.h>
#endif

#ifndef ARPLAYOUT_VIEWSTUBS_H
#include <ArpLayout/ViewStubs.h>
#endif

#ifndef ARPLAYOUT_ARPRUNNINGBAR_H
#include <ArpLayout/ArpRunningBar.h>
#endif

#ifndef PTYREMOTE_H
#include "PTYRemote.h"
#endif

#ifndef PTYREMOTESETTINGS_H
#include "PTYRemoteSettings.h"
#endif

#include <stdlib.h>

ArpMOD();

PTYRemoteSettings::PTYRemoteSettings(
					 const BMessenger& ptyremote,
					 const BMessage& initSettings,
					 const BFont* font,
					 uint32 resizingMode, 
					 uint32 flags)
	: ArpRootLayout(BRect(0,0,100,100), "Shell", resizingMode, flags),
	  mPTYRemote(ptyremote), mSettings(initSettings),
	  mImpl(ptyremote, this, mSettings)
{
	ArpD(cdb << ADH << "PTYRemoteSettings init vals: " << mSettings
					<< endl);
					
	try {
		AddLayoutChild((new ArpRunningBar("TopHBar"))
			->SetParams(ArpMessage()
				.SetInt32(ArpRunningBar::OrientationP, B_HORIZONTAL)
				.SetFloat(ArpRunningBar::IntraSpaceP, .5)
			)
			->AddLayoutChild((new ArpTextControl(
									ArpPTYRemote::CommandConfigName,"Command:","",
									mImpl.AttachTextControl(ArpPTYRemote::CommandConfigName)))
				->SetParams(ArpMessage()
					.SetString(ArpTextControl::MinTextStringP, "WWWWWWWWWW")
					.SetString(ArpTextControl::PrefTextStringP, "WWWWWWWWWWWWWWWWWWWWW")
				)
				->SetConstraints(ArpMessage()
					.SetFloat(ArpRunningBar::WeightC,3)
					.SetInt32(ArpRunningBar::FillC,ArpEastWest)
				)
			)
		);
	} catch(...) {
		throw;
	}
	
	mImpl.RefreshControls(mSettings);
}

PTYRemoteSettings::~PTYRemoteSettings()
{
}

void PTYRemoteSettings::AttachedToWindow()
{
	inherited::AttachedToWindow();
	BView* par = Parent();
	ArpColor col = ui_color(B_PANEL_BACKGROUND_COLOR);
	if( par ) col = par->ViewColor();
	ArpMessage updColor;
	updColor.AddRGBColor("StdBackColor", col);
	UpdateGlobals(updColor);
	
	mImpl.AttachedToWindow();
}

void PTYRemoteSettings::DetachedFromWindow()
{
	inherited::DetachedFromWindow();
	mImpl.DetachedFromWindow();
}

void PTYRemoteSettings::MessageReceived(BMessage *message)
{
	if( mImpl.MessageReceived(message) == B_OK ) return;
	inherited::MessageReceived(message);
}
