/*
	
	SessionSettings.cpp
	
	Copyright (c)1999 by Angry Red Planet.

	This code is distributed under a modified form of the
	Artistic License.  A copy of this license should have
	been included with it; if this wasn't the case, the
	entire package can be obtained at
	<URL:http://www.angryredplanet.com/>.
*/

#ifndef _BUTTON_H
#include <be/interface/Button.h>
#endif

#ifndef _MENUFIELD_H
#include <be/interface/MenuField.h>
#endif

#ifndef _POPUPMENU_H
#include <be/interface/PopUpMenu.h>
#endif

#ifndef _MENUITEM_H
#include <be/interface/MenuItem.h>
#endif

#ifndef _COLORCONTROL_H
#include <be/interface/ColorControl.h>
#endif

#ifndef _STRINGVIEW_H
#include <be/interface/StringView.h>
#endif

#ifndef _CHECKBOX_H
#include <be/interface/CheckBox.h>
#endif

#ifndef _SCREEN_H
#include <be/interface/Screen.h>
#endif

#ifndef _AUTOLOCK_H
#include <be/support/Autolock.h>
#endif

#ifndef _UTF8_H
#include <be/support/UTF8.h>
#endif

#ifndef ARPKERNEL_ARPCOLOR_H
#include <ArpKernel/ArpColor.h>
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include <ArpKernel/ArpDebug.h>
#endif

#ifndef ARPLAYOUT_VIEWSTUBS_H
#include <ArpLayout/ViewStubs.h>
#endif

#ifndef ARPLAYOUT_ARPRUNNINGBAR_H
#include <ArpLayout/ArpRunningBar.h>
#endif

#ifndef TERMINALWIN_H
#include "TerminalWin.h"
#endif

#ifndef SESSIONSETTINGS_H
#include "SessionSettings.h"
#endif

#include <stdlib.h>

ArpMOD();

SessionSettings::SessionSettings(
					 const BMessenger& session,
					 const BMessage& initSettings,
					 const BFont* font,
					 uint32 resizingMode, 
					 uint32 flags)
	: ArpRootLayout(BRect(0,0,100,100), "Session", resizingMode, flags),
	  mSession(session), mSettings(initSettings),
	  mImpl(session, this, mSettings)
{
	ArpD(cdb << ADH << "SessionSettings init vals: " << mSettings
					<< endl);
					
	try {
		AddLayoutChild((new ArpRunningBar("TopVBar"))
			->SetParams(ArpMessage()
				.SetInt32(ArpRunningBar::OrientationP, B_VERTICAL)
				.SetFloat(ArpRunningBar::IntraSpaceP, .5)
			)
			->AddLayoutChild((new ArpRunningBar("TopHBar"))
				->SetParams(ArpMessage()
					.SetInt32(ArpRunningBar::OrientationP, B_HORIZONTAL)
					.SetFloat(ArpRunningBar::IntraSpaceP, .5)
				)
				->SetConstraints(ArpMessage()
					.SetFloat(ArpRunningBar::WeightC,1)
					.SetInt32(ArpRunningBar::FillC,ArpEastWest)
				)
				->AddLayoutChild((new ArpBox("StartupBox","Startup"))
					->SetConstraints(ArpMessage()
						.SetFloat(ArpRunningBar::WeightC,1)
						.SetInt32(ArpRunningBar::FillC,ArpNorthSouth)
					)
					->AddLayoutChild((new ArpRunningBar("StartupVBar"))
						->SetParams(ArpMessage()
							.SetInt32(ArpRunningBar::OrientationP, B_VERTICAL)
							//.SetFloat(ArpRunningBar::IntraSpaceP, .5)
						)
						->AddLayoutChild((new ArpViewWrapper(
								new BCheckBox(BRect(0,0,10,10),
												TerminalWin::StartAtLocationConfigName,
												"Window at Current Location",
												mImpl.AttachCheckBox(TerminalWin::StartAtLocationConfigName),
												B_FOLLOW_NONE,
												B_WILL_DRAW|B_FULL_UPDATE_ON_RESIZE|B_NAVIGABLE) ))
							->SetConstraints(ArpMessage()
								.SetFloat(ArpRunningBar::WeightC,1)
								.SetInt32(ArpRunningBar::FillC,ArpWest)
							)
						)
						->AddLayoutChild((new ArpViewWrapper(
								new BCheckBox(BRect(0,0,10,10),
												TerminalWin::StartAtSizeConfigName,
												"Window at Current Size",
												mImpl.AttachCheckBox(TerminalWin::StartAtSizeConfigName),
												B_FOLLOW_NONE,
												B_WILL_DRAW|B_FULL_UPDATE_ON_RESIZE|B_NAVIGABLE) ))
							->SetConstraints(ArpMessage()
								.SetFloat(ArpRunningBar::WeightC,1)
								.SetInt32(ArpRunningBar::FillC,ArpWest)
							)
						)
						->AddLayoutChild((new ArpViewWrapper(
								new BCheckBox(BRect(0,0,10,10),
												TerminalWin::StartInWorkspaceConfigName,
												"Window in Current Workspace",
												mImpl.AttachCheckBox(TerminalWin::StartInWorkspaceConfigName),
												B_FOLLOW_NONE,
												B_WILL_DRAW|B_FULL_UPDATE_ON_RESIZE|B_NAVIGABLE) ))
							->SetConstraints(ArpMessage()
								.SetFloat(ArpRunningBar::WeightC,1)
								.SetInt32(ArpRunningBar::FillC,ArpWest)
							)
						)
					)
				)
				->AddLayoutChild((new ArpBox("ShutdownBox","Shutdown"))
					->SetConstraints(ArpMessage()
						.SetFloat(ArpRunningBar::WeightC,1)
						.SetInt32(ArpRunningBar::FillC,ArpNorthSouth)
					)
					->AddLayoutChild((new ArpRunningBar("ShutdownVBar"))
						->SetParams(ArpMessage()
							.SetInt32(ArpRunningBar::OrientationP, B_VERTICAL)
							//.SetFloat(ArpRunningBar::IntraSpaceP, .5)
						)
						->AddLayoutChild((new ArpViewWrapper(
								new BCheckBox(BRect(0,0,10,10),
												TerminalWin::CloseConfirmConfigName,
												"Confirm Session Close at Exit",
												mImpl.AttachCheckBox(TerminalWin::CloseConfirmConfigName),
												B_FOLLOW_NONE,
												B_WILL_DRAW|B_FULL_UPDATE_ON_RESIZE|B_NAVIGABLE) ))
							->SetConstraints(ArpMessage()
								.SetFloat(ArpRunningBar::WeightC,1)
								.SetInt32(ArpRunningBar::FillC,ArpWest)
							)
						)
						->AddLayoutChild((new ArpViewWrapper(
								new BCheckBox(BRect(0,0,10,10),
												TerminalWin::ExitOnCloseConfigName,
												"Exit When Session Ends",
												mImpl.AttachCheckBox(TerminalWin::ExitOnCloseConfigName),
												B_FOLLOW_NONE,
												B_WILL_DRAW|B_FULL_UPDATE_ON_RESIZE|B_NAVIGABLE) ))
							->SetConstraints(ArpMessage()
								.SetFloat(ArpRunningBar::WeightC,1)
								.SetInt32(ArpRunningBar::FillC,ArpWest)
							)
						)
						->AddLayoutChild((new ArpViewWrapper(
								new BCheckBox(BRect(0,0,10,10),
												TerminalWin::SaveOnExitConfigName,
												"Automatically Save Session",
												mImpl.AttachCheckBox(TerminalWin::SaveOnExitConfigName,1),
												B_FOLLOW_NONE,
												B_WILL_DRAW|B_FULL_UPDATE_ON_RESIZE|B_NAVIGABLE) ))
							->SetConstraints(ArpMessage()
								.SetFloat(ArpRunningBar::WeightC,1)
								.SetInt32(ArpRunningBar::FillC,ArpWest)
							)
						)
					)
				)
			)
		);
	} catch(...) {
		throw;
	}
	
	mImpl.RefreshControls(mSettings);
}

SessionSettings::~SessionSettings()
{
}

void SessionSettings::AttachedToWindow()
{
	inherited::AttachedToWindow();
	BView* par = Parent();
	ArpColor col = ui_color(B_PANEL_BACKGROUND_COLOR);
	if( par ) col = par->ViewColor();
	ArpMessage updColor;
	updColor.AddRGBColor("StdBackColor", col);
	UpdateGlobals(updColor);
	
	mImpl.AttachedToWindow();
}

void SessionSettings::DetachedFromWindow()
{
	inherited::DetachedFromWindow();
	mImpl.DetachedFromWindow();
}

void SessionSettings::MessageReceived(BMessage *message)
{
	if( message && message->what == ARP_PUT_CONFIGURATION_MSG ) {
		ArpD(cdb << ADH << "Looking for protocol change..." << endl);
		BMessage settings;
		if( message->FindMessage("settings", &settings) == B_OK ) {
			ArpD(cdb << ADH << "Retrieving new protocol..." << endl);
			const char* protocol=0;
			if( settings.FindString(TerminalWin::RemoteProtocolConfigName,
									&protocol) == B_OK && protocol != 0 ) {
				ArpD(cdb << ADH << "Found protocol: " << protocol << endl);
				if( !mCurProtocol.IsEmpty() && mCurProtocol != (const ichar*)protocol ) {
					mImpl.RebuildPanel();
				}
				mCurProtocol = protocol;
			}
		}
	}
	
	if( mImpl.MessageReceived(message) == B_OK ) return;
	inherited::MessageReceived(message);
}
