/*
 * ArpTelnet Copyright (c)1997-98 by Dianne Hackborn.
 * All rights reserved.
 *
 * Based (increasingly loosly) on WebTerm, a Web-based terminal
 * applet written in Java, which is
 * Copyright (C)1996 by the National Alliance for Computational
 * Science and Engineering (NACSE).
 * See <URL:http://www.nacse.org/> for more information.
 *
 * This code is not public domain, nor freely distributable.
 * Please direct any questions or requests to Dianne Hackborn,
 * at <hackbod@lucent.com> or <hackbod@angryredplanet.com>.
 *
 */

#ifndef _AUTOLOCK_H
#include <be/support/Autolock.h>
#endif

#ifndef _APPLICATION_H
#include <be/app/Application.h>
#endif

#ifndef _SCROLL_VIEW_H
#include <be/interface/ScrollView.h>
#endif

#ifndef _MENU_BAR_H
#include <be/interface/MenuBar.h>
#endif

#ifndef _ALERT_H
#include <be/interface/Alert.h>
#endif

#ifndef _SCREEN_H
#include <be/interface/Screen.h>
#endif

#ifndef _UTF8_H
#include <be/support/UTF8.h>
#endif

#ifndef _ROSTER_H
#include <be/app/Roster.h>
#endif

#ifndef _APPFILEINFO_H
#include <be/storage/AppFileInfo.h>
#endif

#ifndef __BSTRING__
#include <be/support/String.h>
#endif

#ifndef _FS_ATTR_H
#include <be/kernel/fs_attr.h>
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include <ArpKernel/ArpDebug.h>
#endif

#ifndef ARPKERNEL_ARPDOCUMENTBUTTON_H
#include <ArpKernel/ArpDocumentButton.h>
#endif

#ifndef ARPTERMINAL_ARPREMOTETERMINAL_H
#include <ArpTerminal/ArpRemoteTerminal.h>
#endif

#ifndef ARPTERMINAL_ARPEMULATOR_H
#include <ArpTerminal/ArpEmulator.h>
#endif

#ifndef ARPNETWORK_ARPTELNET_H
#include <ArpNetwork/ArpTelnet.h>
#endif

#ifndef SETTINGSWIN_H
#include "SettingsWin.h"
#endif

#ifndef SESSIONSETTINGS_H
#include "SessionSettings.h"
#endif

#ifndef PTYREMOTE_H
#include "PTYRemote.h"
#endif

#ifndef TERMINALWIN_H
#include "TerminalWin.h"
#endif

#ifndef MAIN_H
#include "Main.h"
#endif

#include <string.h>

ArpMOD();

/*	------------------------------------------------------------
	PUBLIC CONFIGURATION INFORMATION.
	------------------------------------------------------------ */

const char* TerminalWin::TerminalConfigName = "TerminalConfig";
const char* TerminalWin::RemoteSettingsConfigName = "RemoteConfig";
const char* TerminalWin::RemoteProtocolConfigName = "RemoteName";
const char* TerminalWin::LocationConfigName = "WindowLocation";
const char* TerminalWin::SizeConfigName = "WindowSize";
const char* TerminalWin::WorkspaceConfigName = "WindowWorkspaces";
const char* TerminalWin::StartAtLocationConfigName = "StartAtLocation";
const char* TerminalWin::StartAtSizeConfigName = "StartAtSize";
const char* TerminalWin::StartInWorkspaceConfigName = "StartInWorkspace";
const char* TerminalWin::CloseConfirmConfigName = "CloseConfirmation";
const char* TerminalWin::ExitOnCloseConfigName = "ExitOnClose";
const char* TerminalWin::SaveOnExitConfigName = "SaveOnExit";

// Some of these private messages are implemented, some aren't...
enum {
	QUERY_CONNECT_MSG = '.qcn',
	DISCONNECT_MSG = '.dcn',
	SHOW_MANUAL_MSG = '.smn',
	SET_SETTINGS_MSG = '.set',
	SAVE_SETTINGS_MSG = '.sst',
	ABOUT_REMOTE_MSG = '.arm',
	LOCK_SIZE_MSG = '.lsz',
	RESIZE_TO_FIT_MSG = '.rtf',
	SET_FONT_SIZE_MSG = '.fsz',
	SET_CRLF_MSG = '.scl',
	SET_ENTER_MSG = '.snt',
	SET_SWAPBSDEL_MSG = '.sbd',
	SET_WRAP_MSG = '.swr',
	SET_INVERSE_MSG = '.sin',
	SET_PASTEQUERY_MSG = '.spq',
	SET_RMBPASTE_MSG = '.srp',
	SET_ISCROLL_MSG = '.sis',
	SET_OSCROLL_MSG = '.sos',
	SET_ENCODING_MSG = '.sen',
	SOFT_RESET_MSG = '.srs',
	HARD_RESET_MSG = '.hrs',
	ABOUT_EMULATION_MSG = '.aem',
	DEF_CONFIG_CHANGED_MSG = '.dcc',
	CONFIG_FILE_CHANGED_MSG = '.cfc',
	
	// Something old that we should probably just take out...
	TEST_MY_SCREEN = '.tst',
};

class myMenuBar : public BMenuBar
{
public:
	myMenuBar(	BRect frame,
				const char *title,
				uint32 resizeMask =
					B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP,
				menu_layout layout = B_ITEMS_IN_ROW,
				bool resizeToFit = true)
		: BMenuBar(frame, title, resizeMask, layout, resizeToFit),
		  mLastFrame(Frame())
	{
		SetFlags(Flags()|B_FRAME_EVENTS);
	}
	
	virtual	void FrameResized(float new_width, float new_height) {
		if( Window() ) {
			if( new_width < mLastFrame.Width() ) {
				Invalidate(BRect(new_width-3, Bounds().top,
								 new_width, Bounds().bottom));
			} else if( new_width > mLastFrame.Width() ) {
				Invalidate(BRect(mLastFrame.Width()-3, Bounds().top,
								 mLastFrame.Width(), Bounds().bottom));
			}
		}
		ArpD(cdb << ADH << "myMenuBar: FrameResized(" << new_width
			<< ", " << new_height << ") cur frame=" << Frame() << endl);
		BMenuBar::FrameResized(new_width, new_height);
		ArpD(cdb << ADH << "myMenuBar: FrameResized() new frame=" << Frame() << endl);
		mLastFrame = Frame();
	}
	
private:
	BRect mLastFrame;
};

TerminalWin::TerminalWin(BPoint pos, const char* title,
						 const entry_ref& settings, const char* abouttxt)
	: BWindow(BRect(pos.x,pos.y,pos.x+100,pos.y+100), title,
				B_DOCUMENT_WINDOW, B_ASYNCHRONOUS_CONTROLS ),
	  mEmulators(), mRemotes(), mDefTitle(title), mWatch(this, this),
	  sessionmenu(0), connmenu(0), connitem(0), disconnitem(0),
	  winsizemenu(0), lockitem(0), fontsizemenu(0),
	  encodingmenu(0),
	  crlfitem(0), bsdelitem(0), wrapitem(0), inverseitem(0),
	  pqueryitem(0), rmbpasteitem(0), iscrollitem(0), oscrollitem(0),
	  emumenu(0),
	  terminal(0), remote(0),
	  mDoConnect(false), mHasPosition(false),
	  mStartAtLocation(true), mStartAtSize(true), mStartInWorkspace(true),
	  mCloseConfirm(true), mExitOnClose(false), mSaveOnExit(0)
{
	mDefConfig.SetFile("ARP/ArpTelnet", "Default");
	mDefConfig.SetMimeType(ArpTelnetConfigMime);
	mDefConfig.AddConfig(this);
	
	mConfigFile = mDefConfig;
	UpdateTitle();
	
	mDefConfig.StartWatching(BMessenger(this),
							 new BMessage(DEF_CONFIG_CHANGED_MSG));
	mConfigFile.StartWatching(BMessenger(this),
							  new BMessage(CONFIG_FILE_CHANGED_MSG));
	
	BRect aRect(Bounds());
	aRect.Set(0,0,
			  aRect.Width()-B_V_SCROLL_BAR_WIDTH,
			  aRect.Height()-B_H_SCROLL_BAR_HEIGHT);
	#if 0
	BView* menuView = new BView(Bounds(), "MenuContainer",
								B_FOLLOW_LEFT_RIGHT|B_FOLLOW_TOP, 0);
	if( menuView ) AddChild(menuView);
	#endif
	
	BMenuBar* menu = new myMenuBar(Bounds(),"MainMenu",
								  B_FOLLOW_LEFT_RIGHT|B_FOLLOW_TOP,
								  B_ITEMS_IN_ROW, true);
	BMenu* termmenu = 0;
	
	ArpD(cdb << ADH << "Adding search paths to emulator manager..." << endl);
	
	// Tell the remote manager which directories to look in.
	mRemotes.AddEnvVar("ADDON_PATH","ARP/TermRemote");
	mRemotes.AddDirectory(B_COMMON_ADDONS_DIRECTORY,"ARP/TermRemote");
	mRemotes.AddDirectory(B_USER_ADDONS_DIRECTORY,"ARP/TermRemote");

	ArpD(cdb << ADH << "Looking for remote devices..." << endl);
	
	// Search for all available external add-ons, and add our
	// shell and telnet classes as well.
	mRemotes.Start();
	mRemotes.AddStaticRemote(&ArpTelnet::AddonInterface);
	mRemotes.AddStaticRemote(&ArpPTYRemote::AddonInterface);
	
	ArpD(cdb << ADH << "Constructing remote menu..." << endl);
	
	if( menu ) {
		//if( menuView ) menuView->AddChild(menu);
		AddChild(menu);
		
		sessionmenu = new BMenu("Session");
		if( sessionmenu ) {
			menu->AddItem(sessionmenu);
			
			BMenuItem *item; 
			
			connitem = new BMenuItem("Connect" B_UTF8_ELLIPSIS, 
	                             new BMessage(QUERY_CONNECT_MSG),
	                             'T'); 
			connitem->SetTarget(this); 
			sessionmenu->AddItem(connitem);

			disconnitem = new BMenuItem("Disconnect", 
	                             new BMessage(DISCONNECT_MSG),0);
			disconnitem->SetTarget(this);
			sessionmenu->AddItem(disconnitem);
			
			sessionmenu->AddSeparatorItem();
			
			item = new BMenuItem("Settings" B_UTF8_ELLIPSIS,
	                             new BMessage(SET_SETTINGS_MSG),0); 
			item->SetTarget(this);
			sessionmenu->AddItem(item);

			item = new BMenuItem("Open" B_UTF8_ELLIPSIS,
	                             new BMessage(TERMINALWIN_OPENSETTINGS_MSG),0); 
			item->SetTarget(this);
			sessionmenu->AddItem(item);

			item = new BMenuItem("Save",
	                             new BMessage(SAVE_SETTINGS_MSG),0); 
			item->SetTarget(this);
			sessionmenu->AddItem(item);

			item = new BMenuItem("Save As" B_UTF8_ELLIPSIS,
	                             new BMessage(TERMINALWIN_SAVESETTINGS_MSG),0); 
			item->SetTarget(this);
			sessionmenu->AddItem(item);

			sessionmenu->AddSeparatorItem();
			
			item = new BMenuItem("Help" B_UTF8_ELLIPSIS,
	                             new BMessage(SHOW_MANUAL_MSG)); 
			item->SetTarget(this); 
			sessionmenu->AddItem(item);
	
			entry_ref manref;
			if( GetManualRef(&manref) != B_OK ) {
				item->SetEnabled(false);
			}
			
			item = new BMenuItem(abouttxt,
	                             new BMessage(B_ABOUT_REQUESTED)); 
			item->SetTarget(be_app); 
			sessionmenu->AddItem(item);
	
			item = new BMenuItem("Test Screen", 
	                             new BMessage(TEST_MY_SCREEN)); 
			item->SetTarget(this); 
			sessionmenu->AddItem(item);
			sessionmenu->AddSeparatorItem();
			
			item = new BMenuItem("Quit",
								new BMessage(B_QUIT_REQUESTED),'Q'); 
			//item->SetTarget(be_app); 
			sessionmenu->AddItem(item);
		}
		
		BMenu* edit = new BMenu("Edit");
		if( edit ) {
			BMenuItem *item; 
			
			item = new BMenuItem("Copy",
	                             new BMessage(B_COPY), 'C');
			item->SetTarget(0,this); 
			edit->AddItem(item);

			item = new BMenuItem("Paste", 
	                             new BMessage(B_PASTE), 'V');
			item->SetTarget(0,this); 
			edit->AddItem(item);
			
			item = new BMenuItem("Select All", 
	                             new BMessage(B_SELECT_ALL), 'A'); 
			item->SetTarget(0,this); 
			edit->AddItem(item);
			
			menu->AddItem(edit);
		}
		
		connmenu = new BMenu("Connection");
		if( connmenu ) {
			mRemotes.CreateAddonMenu(connmenu);
			connmenu->SetTargetForItems(this);
			connmenu->AddSeparatorItem();
			
			BMenuItem *item; 
			
			item = new BMenuItem("About Device" B_UTF8_ELLIPSIS, 
	                             new BMessage(ABOUT_REMOTE_MSG));
			item->SetTarget(this); 
			connmenu->AddItem(item);
			
			menu->AddItem(connmenu);
		}
		
		termmenu = new BMenu("Terminal");
		if( termmenu ) menu->AddItem(termmenu);
		
		emumenu = new BMenu("Emulation");
		if( emumenu ) menu->AddItem(emumenu);
		
		float w=0, h=0;
	#if 0
		menu->ResizeTo(1,1);
		menu->Draw(menu->Bounds());
		menu->ResizeToPreferred();
	#endif
		menu->GetPreferredSize(&w,&h);
	#if 0
		menu->ResizeTo(Bounds().Width(), h);
	#endif
		aRect.top += h+1;
		//if( menuView ) menuView->ResizeTo(Bounds().Width(), h);
	}
	
	terminal = new ArpRemoteTerminal(aRect, "Test ArpTerminal");
	if( !terminal ) return;
	
	BScrollView		*scrview;
	scrview = new BScrollView("TerminalScrollView", terminal,
							  B_FOLLOW_ALL_SIDES, 0,
							  true, true, B_NO_BORDER);
	// add view to window
	if( scrview ) AddChild(scrview);
	else AddChild(terminal);

	// insert document button into right side of menu bar
	menu->Draw(menu->Bounds());
	menu->ResizeTo(Bounds().Width()-B_V_SCROLL_BAR_WIDTH+1,
					menu->Bounds().Height());
	ArpD(cdb << ADH << "Win bounds=" << Bounds() << ", menu frame=" << menu->Frame() << endl);
	ArpDocumentButton* dbut = new ArpDocumentButton(
			BRect(Bounds().right-B_V_SCROLL_BAR_WIDTH+1,
				  menu->Bounds().top-1,
				  Bounds().right,
				  menu->Bounds().bottom),
			"DocumentButton", 0, new BMessage(SET_SETTINGS_MSG),
			B_FOLLOW_RIGHT|B_FOLLOW_TOP
		);
	
	if( dbut ) {
		BMessage dropProto(B_SIMPLE_DATA);
		dropProto.AddString("be:types", ArpTelnetConfigMime);
		dropProto.AddString("be:filetypes", ArpTelnetConfigMime);
		dropProto.AddInt32("be:actions", B_COPY_TARGET);
		dbut->SetDropPrototype(dropProto);
		BMessage dragProto(B_SIMPLE_DATA);
		dragProto.AddString("be:types", B_FILE_MIME_TYPE);
		dragProto.AddString("be:types", ArpTelnetConfigMime);
		dragProto.AddString("be:filetypes", ArpTelnetConfigMime);
		dragProto.AddInt32("be:actions", B_COPY_TARGET);
		dragProto.AddInt32("be:actions", B_MOVE_TARGET);
		dragProto.AddString("be:clip_name", "New ArpTelnet Session");
		dbut->SetDragPrototype(dragProto);
		AddChild(dbut);
	}
	
	ArpD(cdb << ADH << "Win bounds=" << Bounds() << endl);
	
	//ArpD(cdb << ADH << "Adding remote device to terminal..." << endl);
	
	terminal->TermSetFixedSize(25,80);
	terminal->TermSetCursorPos(0,0);
	//SetRemote("shell");
	
	ArpD(cdb << ADH << "Locking my window." << endl);
	
	BAutolock wlock(this);
	terminal->MakeFocus(true);

	ArpD(cdb << ADH << "Adding search paths to emulator manager..." << endl);
	
	// Tell the emulator manager which directories to look in.
	mEmulators.AddEnvVar("ADDON_PATH","ARP/TermEmulator");
	mEmulators.AddDirectory(B_COMMON_ADDONS_DIRECTORY,"ARP/TermEmulator");
	mEmulators.AddDirectory(B_USER_ADDONS_DIRECTORY,"ARP/TermEmulator");

#if 0
	// These are the old location of emulator add-ons.  This is only for
	// compatibility -- don't use these.
	mEmulators.AddEnvVar("ADDON_PATH","ArpTerminal");
	mEmulators.AddDirectory(B_COMMON_ADDONS_DIRECTORY,"ArpTerminal");
	mEmulators.AddDirectory(B_USER_ADDONS_DIRECTORY,"ArpTerminal");
#endif

	ArpD(cdb << ADH << "Looking for emulations..." << endl);
	
	// Search for all available external add-ons, and add our
	// ArpEmulator class in case nothing else is there.
	mEmulators.Start();
	mEmulators.AddStaticEmulator(&ArpEmulator::AddonInterface);
	
	ArpD(cdb << ADH << "Constructing menus..." << endl);
	
	// Construct 'Terminal' and 'Emulator' menus.
	if( terminal ) {
		ArpD(cdb << ADH << "Setting terminal to use manager..." << endl);
		terminal->UseEmulatorManager(&mEmulators);
		ArpD(cdb << ADH << "Setting to XTERM emulator..." << endl);
		delete (terminal->SetEmulator("XTERM"));
		if( emumenu ) {
			ArpD(cdb << ADH << "Creating emulator menu..." << endl);
			mEmulators.CreateAddonMenu(emumenu);
			terminal->CheckEmulationMenu(emumenu);
			emumenu->SetTargetForItems(terminal);
			emumenu->AddSeparatorItem();
			BMenuItem* item = new BMenuItem("About Emulator" B_UTF8_ELLIPSIS, 
	                             new BMessage(ABOUT_EMULATION_MSG)); 
			item->SetTarget(0,this); 
			emumenu->AddItem(item);
		}
		if( termmenu ) {
			BMessage* msg;
			BMenuItem* item;
			
			ArpD(cdb << ADH << "Creating terminal menu..." << endl);
			BMenu* winmenu = new BMenu("Usable Size");
			winsizemenu = winmenu;
			if( winmenu ) {
				termmenu->AddItem(winmenu);
				winmenu->SetRadioMode(true);
				msg = new BMessage(TERM_WINDOW_SIZE_MSG);
				if( msg ) {
					msg->AddInt32("rows",24);
					msg->AddInt32("columns",80);
					item = new BMenuItem("80 x 24", msg, '4');
					if( item ) {
						item->SetTarget(terminal);
						winmenu->AddItem(item);
					}
				}
				msg = new BMessage(TERM_WINDOW_SIZE_MSG);
				if( msg ) {
					msg->AddInt32("rows",25);
					msg->AddInt32("columns",80);
					item = new BMenuItem("80 x 25", msg, '5');
					if( item ) {
						item->SetTarget(terminal);
						winmenu->AddItem(item);
						item->SetMarked(true);
					}
				}
				msg = new BMessage(TERM_WINDOW_SIZE_MSG);
				if( msg ) {
					msg->AddInt32("rows",-1);
					msg->AddInt32("columns",80);
					item = new BMenuItem("80 Columns", msg, '8');
					if( item ) {
						item->SetTarget(terminal);
						winmenu->AddItem(item);
					}
				}
				msg = new BMessage(TERM_WINDOW_SIZE_MSG);
				if( msg ) {
					msg->AddInt32("rows",24);
					msg->AddInt32("columns",132);
					item = new BMenuItem("132 x 24", msg);
					if( item ) {
						item->SetTarget(terminal);
						winmenu->AddItem(item);
					}
				}
				msg = new BMessage(TERM_WINDOW_SIZE_MSG);
				if( msg ) {
					msg->AddInt32("rows",25);
					msg->AddInt32("columns",132);
					item = new BMenuItem("132 x 25", msg);
					if( item ) {
						item->SetTarget(terminal);
						winmenu->AddItem(item);
					}
				}
				msg = new BMessage(TERM_WINDOW_SIZE_MSG);
				if( msg ) {
					msg->AddInt32("rows",-1);
					msg->AddInt32("columns",132);
					item = new BMenuItem("132 Columns", msg);
					if( item ) {
						item->SetTarget(terminal);
						winmenu->AddItem(item);
					}
				}
				msg = new BMessage(LOCK_SIZE_MSG);
				if( msg ) {
					lockitem = new BMenuItem("Lock At ?", msg, 'L');
					if( lockitem ) {
						lockitem->SetTarget(this);
						winmenu->AddItem(lockitem);
					}
				}
				msg = new BMessage(TERM_WINDOW_SIZE_MSG);
				if( msg ) {
					msg->AddInt32("rows",-1);
					msg->AddInt32("columns",-1);
					item = new BMenuItem("Size With Window", msg, 'S');
					if( item ) {
						item->SetTarget(terminal);
						winmenu->AddItem(item);
					}
				}
			}
			
			BMenu* fsizemenu = new BMenu("Font Size");
			fontsizemenu = fsizemenu;
			if( fsizemenu ) {
				termmenu->AddItem(fsizemenu);
				fsizemenu->SetRadioMode(true);
				BFont font;
				if( terminal ) {
					terminal->GetFont(&font);
				}
				static const int32 sizes[] = {
					1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14,
					16, 18, 20, 24, 28, 32, 40, 60, 80, 120, 0 };
				for( int32 i=0; sizes[i]; i++ ) {
					msg = new BMessage(SET_FONT_SIZE_MSG);
					if( msg ) {
						msg->AddInt32("size",sizes[i]);
						char buffer[32];
						sprintf(&buffer[0],"%ld",sizes[i]);
						item = new BMenuItem(&buffer[0], msg);
						if( item ) {
							item->SetTarget(this);
							fsizemenu->AddItem(item);
							if( sizes[i] == font.Size() ) {
								item->SetMarked(true);
							}
						}
					}
				}
			}
			
			BMenu* encmenu = new BMenu("Encoding");
			encodingmenu = encmenu;
			if( encmenu ) {
				termmenu->AddItem(encmenu);
				encmenu->SetRadioMode(true);
				static const char* encnames[] = {
					"ISO 8859 Latin 1",
					"ISO 8859 Latin 2",
					"ISO 8859 Latin 3",
					"ISO 8859 Latin 4",
					"ISO 8859 Latin/Cyrillic",
					"ISO 8859 Latin/Arabic",
					"ISO 8859 Latin/Hebrew",
					"ISO 8859 Latin 5",
					"ISO 8859 Latin 6",
					"Macintosh Roman",
					0
				};
				static int32 enccodes[] = {
					B_ISO1_CONVERSION,
					B_ISO2_CONVERSION,
					B_ISO3_CONVERSION,
					B_ISO4_CONVERSION,
					B_ISO5_CONVERSION,
					B_ISO6_CONVERSION,
					B_ISO7_CONVERSION,
					B_ISO8_CONVERSION,
					B_ISO9_CONVERSION,
					B_ISO10_CONVERSION,
					B_MAC_ROMAN_CONVERSION 
				};
				for( int32 i=0; encnames[i] != 0; i++ ) {
					msg = new BMessage(SET_ENCODING_MSG);
					if( msg ) {
						msg->AddInt32("encoding",enccodes[i]);
						item = new BMenuItem(encnames[i], msg);
						if( item ) {
							item->SetTarget(this);
							encmenu->AddItem(item);
							if( enccodes[i] == terminal->TermEncoding() ) {
								item->SetMarked(true);
							}
						}
					}
				}
			}
			
			BMenu* modemenu = new BMenu("Operation Modes");
			if( modemenu ) {
				termmenu->AddItem(modemenu);
				crlfitem = new BMenuItem("Receive LF as CRLF",
								new BMessage(SET_CRLF_MSG),'F');
				if( crlfitem ) {
					crlfitem->SetTarget(this);
					modemenu->AddItem(crlfitem);
				}
				enteritem = new BMenuItem("Send Enter as CRLF",
								new BMessage(SET_ENTER_MSG),'E');
				if( enteritem ) {
					enteritem->SetTarget(this);
					modemenu->AddItem(enteritem);
				}
				bsdelitem = new BMenuItem("Swap Backspace/Del",
								new BMessage(SET_SWAPBSDEL_MSG),'B');
				if( bsdelitem ) {
					bsdelitem->SetTarget(this);
					modemenu->AddItem(bsdelitem);
				}
				wrapitem = new BMenuItem("Wrap Cursor",
								new BMessage(SET_WRAP_MSG),'P');
				if( wrapitem ) {
					wrapitem->SetTarget(this);
					modemenu->AddItem(wrapitem);
				}
				inverseitem = new BMenuItem("Inverse",
								new BMessage(SET_INVERSE_MSG),'I');
				if( inverseitem ) {
					inverseitem->SetTarget(this);
					modemenu->AddItem(inverseitem);
				}
				
				modemenu->AddSeparatorItem();
			
				pqueryitem = new BMenuItem("Verify Paste",
								new BMessage(SET_PASTEQUERY_MSG),0);
				if( pqueryitem ) {
					pqueryitem->SetTarget(this);
					modemenu->AddItem(pqueryitem);
				}
				rmbpasteitem = new BMenuItem("Quick Paste",
								new BMessage(SET_RMBPASTE_MSG),0);
				if( rmbpasteitem ) {
					rmbpasteitem->SetTarget(this);
					modemenu->AddItem(rmbpasteitem);
				}
				iscrollitem = new BMenuItem("Scroll on Input",
								new BMessage(SET_ISCROLL_MSG),0);
				if( iscrollitem ) {
					iscrollitem->SetTarget(this);
					modemenu->AddItem(iscrollitem);
				}
				oscrollitem = new BMenuItem("Scroll on Output",
								new BMessage(SET_OSCROLL_MSG),0);
				if( oscrollitem ) {
					oscrollitem->SetTarget(this);
					modemenu->AddItem(oscrollitem);
				}
				update_modes();
			}

			termmenu->AddSeparatorItem();
			
			item = new BMenuItem("Soft Reset", 
	                             new BMessage(SOFT_RESET_MSG),'R'); 
			if( item ) {
				item->SetTarget(this);
				termmenu->AddItem(item);
			}
			
			item = new BMenuItem("Hard Reset", 
	                             new BMessage(HARD_RESET_MSG),'H'); 
			if( item ) {
				item->SetTarget(this);
				termmenu->AddItem(item);
			}
		
			item = new BMenuItem("Resize To Fit", 
	                             new BMessage(RESIZE_TO_FIT_MSG),'Y'); 
			if( item ) {
				item->SetTarget(this);
				termmenu->AddItem(item);
			}
		}
	}
	
	// An unsuccessful attempt to dispatch cursor keys with
	// command down to the terminal...
#if 0
	BMessage* msg = new BMessage(B_KEY_DOWN);
	if( msg ) {
		msg->AddInt32("modifiers",B_COMMAND_KEY|B_OPTION_KEY);
		msg->AddInt32("raw_char",B_UP_ARROW);
		msg->AddInt32("byte",B_UP_ARROW);
		msg->AddString("bytes","");
		AddShortcut(B_UP_ARROW,0,msg,terminal);
	}
#endif

	ArpD(cdb << ADH << "Setting window size..." << endl);
	// Snap window to correct size and place cursor at top/left.
	ResizeToTerm();
	ArpD(cdb << ADH << "Done constructing terminal window!" << endl);
	
	if( settings != entry_ref() ) OpenConfiguration(settings);
}

TerminalWin::~TerminalWin()
{
	//cerr << "Current settings:" << endl;
	BMessage settings;
	if( terminal ) {
		terminal->GetConfiguration(&settings);
	}
	//cerr << settings << endl;
	
	BMessage msg(B_QUIT_REQUESTED);
	if( mSettings.IsValid() ) mSettings.SendMessage(&msg);
	
	if( remote ) remote->Delete();
}

void TerminalWin::UpdateTitle()
{
	BAutolock l(this);
	if( !l.IsLocked() ) return;
	
	if( mConfigFile.HasFile() && mConfigFile.EntryRef() != mDefConfig.EntryRef() ) {
		BString title("[");
		title << mConfigFile.EntryRef().name;
		title << "]";
		if( mTermTitle != "" ) {
			title << " ";
			title << mTermTitle;
		}
		SetTitle(title.String());
	} else if( mTermTitle != "" ) {
		SetTitle(mTermTitle.String());
	} else {
		SetTitle(mDefTitle.String());
	}
}

status_t TerminalWin::GetManualRef(entry_ref* ref)
{
	status_t err;
	app_info ai;
	if( (err=be_app->GetAppInfo(&ai)) != 0 ) {
		return err;
	}
	BEntry entry(&ai.ref);
	if( (err=entry.InitCheck()) != 0 ) {
		return err;
	}
	BDirectory dir;
	if( (err=entry.GetParent(&dir)) != 0 ) {
		return err;
	}
	if( (err=dir.FindEntry("Manual/index.html", &entry)) != 0 ) {
		return err;
	}
	if( ref ) err = entry.GetRef(ref);
	return err;
}

status_t TerminalWin::OpenConfiguration(const entry_ref& ref, bool connect)
{
	BAutolock l(this);
	
	status_t err = B_OK;
	BString url;
	
	ArpD(cdb << ADH << "TerminalWin: Opening new configuration file." << endl);
	
	mHasPosition = false;
	
	if( ref != entry_ref() ) {
		ArpD(cdb << ADH << "From file entry: device=" << ref.device
					<< ", dir=" << ref.directory
					<< ", name=" << ref.name << endl);
		BFile file(&ref, B_READ_ONLY);
		status_t err = file.InitCheck();
		if( !err ) {
			ArpD(cdb << ADH << "Created file." << endl);
			attr_info attrinfo;
			if( file.GetAttrInfo("META:url", &attrinfo) == B_OK ) {
				ArpD(cdb << ADH << "Found META:url attribute, type="
							<< attrinfo.type << ", size="
							<< attrinfo.size << endl);
				if( attrinfo.type == B_STRING_TYPE ) {
					ssize_t ret = file.ReadAttr("META:url", B_STRING_TYPE, 0,
												url.LockBuffer(attrinfo.size+4),
												attrinfo.size+4);
					ArpD(cdb << ADH << "Read return = " << ret << endl);
					if( ret >= 0 ) {
						url.UnlockBuffer(ret);
						ArpD(cdb << ADH << "Read URL: " << url.String() << endl);
					} else {
						url.UnlockBuffer(0);
						ArpD(cdb << ADH << "Error: " << strerror(ret) << endl);
					}
				}
			}
		}
		file.Unset();
		
		mConfigFile.SetFile(ref);
	}
	
	if( !mConfigFile.HasFile() ) {
		mConfigFile = mDefConfig;
	}
	
	mDoConnect = url != "" ? false : connect;

	err = mConfigFile.ReadSettings();
	if( err && url != "" ) {
		// If unable to read file and the file being read has
		// a URL attribute, instead read the default configuration.
		mConfigFile = mDefConfig;
		err = mConfigFile.ReadSettings();
	}
	
	// If URL specified, use that remote device instead of the
	// one read into the configuration.
	if( url != "" ) {
		SetRemote(url.String(), false);
		mTermTitle = "";
		if( Remote() ) Remote()->Connect(url.String(), this);
	}
	
	UpdateTitle();
	
	mDoConnect = false;
	return err;
}

status_t TerminalWin::SaveConfiguration(entry_ref dir, const char* name,
									    bool showError, bool moveTo)
{
	BAutolock l(this);
	status_t err = B_OK;
	
	ArpConfigureFile* saveConfig = &mConfigFile;
	ArpConfigureFile tempConfig;
	if( !moveTo ) {
		tempConfig = mConfigFile;
		saveConfig = &tempConfig;
	}
	
	if( dir != entry_ref() ) {
		BEntry entry(&dir);
		err = entry.InitCheck();
		BPath path;
		if( !err ) {
			entry.GetPath(&path);
			err = path.InitCheck();
		}
		if( !err && name ) err = path.Append(name);
		if( !err ) saveConfig->SetFile(path);
	}
	
	UpdateTitle();
	
	if( !err ) err = saveConfig->WriteSettings();
	
	if( showError && err ) {
		BString msg("An error occured while writing the file\n");
		msg += saveConfig->Path().Path();
		msg += ":\n";
		msg += strerror(err);
		BAlert* alert = new BAlert("Error Writing Settings",
									msg.String(), "Dismiss");
		if( alert ) {
			alert->Go(0);
		}
	}
	
	UpdateTitle();
	
	return err;
}

status_t TerminalWin::GetConfiguration(BMessage* values) const
{
	status_t res=B_OK;
	
	if( !values ) return B_BAD_VALUE;
	
	BAutolock l(const_cast<TerminalWin*>(this));
	if( !l.IsLocked() ) return B_ERROR;
	
	if( res == B_OK && terminal ) {
		BMessage msg;
		res = terminal->GetConfiguration(&msg);
		if( res == B_OK ) {
			values->AddMessage(TerminalConfigName, &msg);
		}
	}
	
	if( res == B_OK && Remote() ) {
		BMessage msg;
		const ArpRemoteType* type = Remote()->RemoteType();
		res = Remote()->GetConfiguration(&msg);
		if( type && res == B_OK ) {
			res = values->AddMessage(RemoteSettingsConfigName, &msg);
			if( res == B_OK ) {
				res = values->AddString(RemoteProtocolConfigName, type->URL);
			}
		}
	}
	
	if( res == B_OK ) {
		res = GetThisConfiguration(values);
	}
		
	return res;
}

status_t TerminalWin::GetThisConfiguration(BMessage* values) const
{
	status_t res=B_OK;
	
	if( !values ) return B_BAD_VALUE;
	
	BAutolock l(const_cast<TerminalWin*>(this));
	if( !l.IsLocked() ) return B_ERROR;
	
	const bool locked = const_cast<TerminalWin*>(this)->Lock();
	
	try {
		if( res == B_OK && mStartAtLocation ) {
			res = values->AddPoint(LocationConfigName,
								   BPoint(Frame().left,Frame().top));
		}
		
		if( res == B_OK ) {
			res = values->AddBool(StartAtLocationConfigName, mStartAtLocation);
		}
		
		if( res == B_OK && terminal && mStartAtSize ) {
			int32 rows=0, cols=0;
			terminal->TermGetSize(&rows, &cols);
			res = values->AddPoint(SizeConfigName, BPoint(rows, cols));
		}
		
		if( res == B_OK ) {
			res = values->AddBool(StartAtSizeConfigName, mStartAtSize);
		}
		
		if( res == B_OK && mStartInWorkspace ) {
			res = values->AddInt32(WorkspaceConfigName, (int32)Workspaces());
		}
		
		if( res == B_OK ) {
			res = values->AddBool(StartInWorkspaceConfigName, mStartInWorkspace);
		}
		
		if( res == B_OK ) {
			res = values->AddBool(CloseConfirmConfigName, mCloseConfirm);
		}
		
		if( res == B_OK ) {
			res = values->AddBool(ExitOnCloseConfigName, mExitOnClose);
		}
		
		if( res == B_OK ) {
			res = values->AddInt32(SaveOnExitConfigName, mSaveOnExit);
		}
		
	} catch(...) {
		if( locked ) const_cast<TerminalWin*>(this)->Unlock();
		throw;
	}
	
	if( locked ) const_cast<TerminalWin*>(this)->Unlock();
	
	return res;
}

status_t TerminalWin::PutConfiguration(const BMessage* values)
{
	status_t res = B_OK;
	
	if( !values ) return B_BAD_VALUE;
	
	const bool locked = Lock();
	mHasPosition = false;
	
	try {
		if( res == B_OK && values->HasString(RemoteProtocolConfigName) ) {
			const char* protocol = 0;
			res = values->FindString(RemoteProtocolConfigName, &protocol);
			if( res == B_OK ) {
				const ArpRemoteType* type = Remote() ? Remote()->RemoteType() : 0;
				// Change remote if different...
				if( !type || strcmp(type->URL, protocol) ) {
					ArpString url(protocol);
					url += ":";
					res = SetRemote(url, false);
					if( !Remote() ) res = B_ERROR;
				}
			}
		}
		if( res == B_OK && Remote() &&
				values->HasMessage(RemoteSettingsConfigName) ) {
			BMessage msg;
			res = values->FindMessage(RemoteSettingsConfigName, &msg);
			if( res == B_OK ) res = Remote()->PutConfiguration(&msg);
		}
		if( res == B_OK && terminal &&
				values->HasMessage(TerminalConfigName) ) {
			BMessage msg;
			res = values->FindMessage(TerminalConfigName, &msg);
			if( res == B_OK ) res = terminal->PutConfiguration(&msg);
		}
		
		if( res == B_OK && values->HasBool(StartInWorkspaceConfigName) ) {
			res = values->FindBool(StartInWorkspaceConfigName,
								 &mStartInWorkspace);
		}
		
		if( res == B_OK && mStartInWorkspace &&
				values->HasInt32(WorkspaceConfigName) ) {
			int32 ws=0;
			if( values->FindInt32(WorkspaceConfigName, &ws) == B_OK
						&& ws != (int32)Workspaces() ) {
				SetWorkspaces((uint32)ws);
				if( (ws&(1<<current_workspace())) == 0 ) {
					int index=0;
					while( (ws&1) == 0 ) {
						index++;
						ws >>= 1;
					}
					activate_workspace(index);
				}
			}
		}
		
		if( res == B_OK && values->HasBool(CloseConfirmConfigName) ) {
			res= values->FindBool(CloseConfirmConfigName, &mCloseConfirm);
		}
		
		if( res == B_OK && values->HasBool(ExitOnCloseConfigName) ) {
			res= values->FindBool(ExitOnCloseConfigName, &mExitOnClose);
		}
		
		if( res == B_OK && values->HasInt32(SaveOnExitConfigName) ) {
			res= values->FindInt32(SaveOnExitConfigName, &mSaveOnExit);
		}
		
		if( res == B_OK && values->HasBool(StartAtLocationConfigName) ) {
			res = values->FindBool(StartAtLocationConfigName,
								 &mStartAtLocation);
		}
		
		if( res == B_OK && values->HasBool(StartAtSizeConfigName) ) {
			res = values->FindBool(StartAtSizeConfigName, &mStartAtSize);
		}
		
		BScreen screen(this);
		if( screen.IsValid() && res == B_OK ) {
			
			BRect sfrm = screen.Frame();
			float w=Frame().Width(), h=Frame().Height();
			BPoint size;
			if( mStartAtSize == true
					&& values->FindPoint(SizeConfigName,&size) == B_OK ) {
				ResizeToTerm((int32)size.x, (int32)size.y);
				if( w > sfrm.Width() || h > sfrm.Height() ) {
					w = w > sfrm.Width() ? sfrm.Width() : w;
					h = h > sfrm.Height() ? sfrm.Height() : h;
					ResizeTo(w, h);
				}
			}
			
			BPoint loc;
			if( mStartAtLocation == true
					&& values->FindPoint(LocationConfigName,&loc) == B_OK ) {
				mHasPosition = true;
				if( loc.x < (-w+10) ) loc.x = -w+10;
				else if( loc.x > (sfrm.Width()-10) ) {
					loc.x = sfrm.Width() - 10;
				}
				if( loc.y < (-h+10) ) loc.y = -h+10;
				else if( loc.y > (sfrm.Height()-10) ) {
					loc.y = sfrm.Height() - 10;
				}
				MoveTo(loc);
			}
		}
		
	} catch(...) {
		if( locked ) Unlock();
		throw;
	}
	
	if( res == B_OK ) mWatch.ReportChange(values);
	
	if( locked ) Unlock();
	
	if( res == B_OK && Remote() && !Remote()->IsConnected() && mDoConnect ) {
		mTermTitle = "";
		Remote()->Connect(0, this);
	}
	
	UpdateTitle();
	
	return res;
}

status_t TerminalWin::Configure(ArpVectorI<BView*>& views)
{
	status_t res = B_OK;
	
	BAutolock l(this);
	if( !l.IsLocked() ) return B_ERROR;
	
	BMessage settings;
	if( (res=GetThisConfiguration(&settings)) != B_NO_ERROR ) {
		ArpD(cdb << ADH << "Error getting settings: " << res << endl);
		return res;
	}
	
	views.push_back(new SessionSettings(BMessenger(this), settings));
	
	if( terminal ) res = terminal->Configure(views);
	if( res == B_OK && Remote() ) res = Remote()->Configure(views);
	
	return res;
}

void TerminalWin::update_modes(void)
{
	if( Remote() ) {
		if( Remote()->IsConnected() ) {
			connitem->SetEnabled(false);
			disconnitem->SetEnabled(true);
		} else {
			connitem->SetEnabled(true);
			disconnitem->SetEnabled(false);
		}
	} else {
		connitem->SetEnabled(false);
		disconnitem->SetEnabled(false);
	}
	
	if( terminal ) {
		int32 mode = terminal->TermMode();
		if( crlfitem )
			crlfitem->SetMarked(mode&terminal->TERM_MODENEWLINE);
		if( enteritem )
			enteritem->SetMarked(!(mode&terminal->TERM_MODEENTERISLF));
		if( bsdelitem )
			bsdelitem->SetMarked(mode&terminal->TERM_MODESWAPBSDEL);
		if( wrapitem )
			wrapitem->SetMarked(!(mode&terminal->TERM_MODENOWRAP));
		if( inverseitem )
			inverseitem->SetMarked(mode&terminal->TERM_MODEINVERSE);
		if( pqueryitem )
			pqueryitem->SetMarked(terminal->PasteVerified());
		if( rmbpasteitem )
			rmbpasteitem->SetMarked(terminal->RMBPaste());
		if( iscrollitem )
			iscrollitem->SetMarked(((int)terminal->AutoScrollMode())
									& terminal->AUTOSCROLL_INPUT);
		if( oscrollitem )
			oscrollitem->SetMarked(((int)terminal->AutoScrollMode())
									& terminal->AUTOSCROLL_OUTPUT);
	}
}

void TerminalWin::ClearRemote()
{
	BAutolock l(this);
	if( remote ) {
		ArpD(cdb << ADH << "Deleting old remote device." << endl);
		remote->Delete();
		remote = 0;
		mTermTitle = "";
		UpdateTitle();
	}
}

status_t TerminalWin::SetRemote(const char* url, bool connect)
{
	BAutolock l(this);
	
	if( !url ) return B_ERROR;
	
	ArpD(cdb << ADH << "Setting remote device: url=" << url << endl);
	
	if( !Lock() ) return B_ERROR;
	
	try {
		ArpString protocol;
		ArpParseURL(ArpString(url), &protocol, 0, 0, 0, 0);
		
		if( protocol.IsNull() ) {
			Unlock();
			return B_ERROR;
		}
		
		ArpD(cdb << ADH << "Protocol is " << protocol << endl);
		
		if( !Remote() || protocol != ArpLIT(Remote()->RemoteType()->URL) ) {
			if( Remote() && Remote()->IsInUse() ) {
				ArpD(cdb << ADH << "Asking user to replace current session." << endl);
				ArpString msg("This will replace the current ");
				msg += Remote()->RemoteType()->LongName;
				msg += " session.\nClose it?";
				BAlert* alert = new BAlert("Close Session", msg, "Cancel", "Close Session",
											0, B_WIDTH_AS_USUAL, B_WARNING_ALERT);
				if( !alert || alert->Go() != 1 ) return B_ERROR;
			}
			ClearRemote();
			//terminal->TermClear();
			//terminal->TermClean();
			ArpD(cdb << ADH << "Allocating new remote device." << endl);
			remote = mRemotes.AllocRemote(protocol);
			if( Remote() ) {
				ArpD(cdb << ADH << "Attaching terminal to remote device." << endl);
				terminal->SetRemote(Remote()->GetDevice());
				if( connect ) {
					ArpD(cdb << ADH << "Connecting to remote device." << endl);
					Remote()->Connect(url, this);
				}
			}
			CheckRemoteMenu();
			
			// Temporary hack: inform any active configuration views about
			// the change in remote, so they can rebuild their panel.
			BMessage settings;
			const ArpRemoteType* type = Remote()->RemoteType();
			if( settings.AddString(RemoteProtocolConfigName, type->URL) == B_OK ) {
				mWatch.ReportChange(&settings);
			}
		}
	} catch( ... ) {
		Unlock();
		throw;
	}
	
	Unlock();
	return B_OK;
}

bool TerminalWin::CheckRemoteMenu()
{
	if( !connmenu ) return false;
	
	const ArpRemoteType* type = 0;
	if( Remote() ) type = Remote()->RemoteType();
	bool okay = false;
	for( int i=0; i<connmenu->CountItems(); i++ ) {
		BMenuItem* item = connmenu->ItemAt(i);
		if( item ) {
			BMessage* msg = item->Message();
			if( msg && msg->what == TERM_SELECT_REMOTE_MSG ) {
				const char* str;
				if( !msg->FindString("url",0,&str) ) {
					if( type && !strcmp(type->URL,str) ) {
						item->SetMarked(true);
						okay = true;
					} else item->SetMarked(false);
				}
			}
		}
	}
	return okay;
}

void TerminalWin::PrintToTerm(ArpString str, int32 flags)
{
	BAutolock l(this);
	if( terminal ) terminal->TermSendTTY((const ichar*)str,
											str.Length(),flags);
}

void TerminalWin::Zoom(BPoint leftTop, float width, float height)
{
	ResizeToTerm();
}

void TerminalWin::ResizeToTerm(int32 rows, int32 cols)
{
	BAutolock l(this);
	if( terminal ) {
		terminal->TermGetSize(rows < 0 ? &rows:0,
								cols < 0 ? &cols:0);
		float w,h;
		terminal->GetNeededSize(cols,rows,&w,&h);
		BRect win = Frame();
		BRect inr = terminal->Frame();
		w = w + floor(win.Width()-inr.Width())-1;
		h = h + floor(win.Height()-inr.Height())-1;
		ResizeTo(w,h);
		// For some reason, the terminal is not getting updated
		// corrently when we manually set the window size...
		// (It's only drawing the newly exposed regions.)
		// This is a temporary hack to fix this.
		terminal->Invalidate();
	}
}

void TerminalWin::DispatchMessage(BMessage *message,
								  BHandler *handler)
{
	ArpD(if( message->what != B_MOUSE_MOVED )
				cdb << ADH << "Dispatch message to " << handler
					<< ": " << *message << endl);
	if( message && message->what == B_KEY_DOWN || message->what == B_KEY_UP ) {
		BView* view = dynamic_cast<BView*>(handler);
		if( view ) {
			// HACK HACK HACK...  when using non-unicode encodings, CTRL-space doesn't
			// seem to get delivered to the view...  so force it to get there.
			int32 mods = 0;
			int32 key = 0;
			message->FindInt32("modifiers",&mods);
			message->FindInt32("raw_char",&key);
			if( (key == ' ' || key == '2') && (mods&B_CONTROL_KEY) != 0 ) {
				char byte = 0;
				if( message->what == B_KEY_DOWN ) view->KeyDown(&byte,1);
				else view->KeyUp(&byte,1);
				return;
			}
		}
	}
	inherited::DispatchMessage(message,handler);
}

void TerminalWin::MessageReceived(BMessage* message)
{
	if( !message ) return;
	
	if( mWatch.MessageReceived(message) == B_OK ) return;
	
	switch( message->what ) {
		case DEF_CONFIG_CHANGED_MSG:
		case CONFIG_FILE_CHANGED_MSG: {
			// If the node monitor on our configuration file changes, update
			// the window title.
			UpdateTitle();
		} break;
		case QUERY_CONNECT_MSG: {
			// If the user selects the connect menu, have the remote device
			// do whatever its default behavior is.
			if( Remote() ) Remote()->Connect(this);
		} break;
		case ARPTELNET_CONNECT_MSG:
		case B_REFS_RECEIVED: {
			// Forward connect requests to the remote device.
			if( message->HasRef("refs") ) {
				entry_ref ref;
				if( message->FindRef("refs", &ref) == B_OK ) {
					OpenConfiguration(ref, true);
					return;
				}
			}
			if( message->what == ARPTELNET_CONNECT_MSG && Remote() ) {
				Remote()->GetDevice().SendMessage(message);
			}
		} break;
		case DISCONNECT_MSG: {
			if( Remote() ) Remote()->Disconnect();
		} break;
		case SHOW_MANUAL_MSG: {
			entry_ref ref;
			if( GetManualRef(&ref) == B_OK ) {
				be_roster->Launch(&ref);
			}
		} break;
		case SET_SETTINGS_MSG: {
			if( terminal /*&& !mSettings.IsValid()*/ ) {
				//ArpConfigureFile defConfig(mConfigFile);
				//defConfig.SetFile("ARP/ArpTelnet", "Default");
				SettingsWin* sw = new SettingsWin("ArpTelnet Settings",
												  mDefConfig, this);
				sw->Show();
				mSettings = BMessenger(sw);
			}
		} break;
		case TERMINALWIN_OPENSETTINGS_MSG: {
			BMessage msg(TERMINALWIN_OPENSETTINGS_MSG);
			status_t err = B_OK;
			if( mConfigFile.HasFile() ) {
				entry_ref ref = mConfigFile.EntryRef();
				BEntry entry(&ref);
				if( entry.InitCheck() == B_OK ) {
					entry_ref ref;
					err = entry.GetRef(&ref);
					if( !err ) err = msg.AddRef("current", &ref);
				}
			}
			if( !err ) err = msg.AddMessenger("who", BMessenger(this));
			if( !err ) {
				be_app_messenger.SendMessage(&msg, this);
			}
		} break;
		case B_SAVE_REQUESTED:
		case B_MOVE_TARGET:
		case B_COPY_TARGET: {
			ArpD(cdb << ADH << "Move/Save from message: " << *message << endl);
			entry_ref dir;
			const char* name;
			status_t err = message->FindRef("directory", &dir);
			if( !err ) err = message->FindString("name", &name);
			if( !err ) {
				// Asked to write into a file.
				ArpD(cdb << ADH << "Saving into file." << endl);
				SaveConfiguration(dir, name, true,
								  message->what==B_COPY_TARGET ? false:true);
			} else {
				// Asked to reply with data.
				ArpD(cdb << ADH << "Replying in message." << endl);
				const char* type=0;
				err = message->FindString("be:types", &type);
				if( !err && strcasecmp(ArpTelnetConfigMime, type)!=0 ) {
					err = B_ERROR;
				}
				BMallocIO f;
				if( !err ) {
					BMessage data;
					err = mConfigFile.MakeSettings(&data);
					if( !err ) err = data.Flatten(&f);
				}
				if( !err ) {
					BMessage reply(B_MIME_DATA);
					reply.AddData(ArpTelnetConfigMime, B_MIME_TYPE,
								  f.Buffer(), f.BufferLength());
					ArpD(cdb << ADH << "Reply is: " << reply << endl);
					message->SendReply(&reply);
				}
			}
		} break;
		case B_MIME_DATA: {
			ArpD(cdb << ADH << "Loading from message: " << *message << endl);
			const void* data=0;
			ssize_t size=0;
			status_t err;
			err = message->FindData(ArpTelnetConfigMime, B_MIME_TYPE, &data, &size);
			ArpD(cdb << ADH << "Status=" << err << ", size=" << size << endl);
			if( !err ) {
				BMemoryIO memio(data, size);
				BMessage settings;
				err = settings.Unflatten(&memio);
				ArpD(cdb << ADH << "Settings are: " << settings << endl);
				if( !err ) {
					mDoConnect = true;
					ArpD(cdb << ADH << "Applying settings." << endl);
					err = mConfigFile.ApplySettings(&settings);
				}
				UpdateTitle();
				mDoConnect = false;
			}
		} break;
		case SAVE_SETTINGS_MSG: {
			if( mConfigFile.HasFile() ) {
				SaveConfiguration();
				break;
			}
			
			// If no settings file is currently set, fall through to bring
			// up "save" requester.
		}
		case TERMINALWIN_SAVESETTINGS_MSG: {
			BMessage msg(TERMINALWIN_SAVESETTINGS_MSG);
			status_t err = B_OK;
			if( mConfigFile.HasFile() ) {
				entry_ref ref = mConfigFile.EntryRef();
				BEntry entry(&ref);
				if( entry.InitCheck() == B_OK ) {
					entry_ref ref;
					err = entry.GetRef(&ref);
					if( !err ) err = msg.AddRef("current", &ref);
				}
			}
			if( !err ) err = msg.AddMessenger("who", BMessenger(this));
			if( !err ) {
				be_app_messenger.SendMessage(&msg, this);
			}
		} break;
		case B_PASTE: {
			// Just used for my own testing purposes...
			#if defined(ArpDEBUG)
			if(  be_clipboard->Lock() ) {
				BMessage* clip = be_clipboard->Data();
				ArpD(cdb << ADH << "Clipboard data: " << *clip
						<< endl);
				be_clipboard->Unlock();
			}
			#endif
			inherited::MessageReceived(message);
		} break;
		case TERM_SELECT_REMOTE_MSG: {
			const char* protocol;
			if( message->FindString("url", &protocol) == B_NO_ERROR ) {
				ArpString url(protocol);
				url += ArpLIT(":");
				SetRemote(url);
			}
		} break;
		case TERM_SET_TITLE_MSG: {
			const char* title;
			if( message->FindString("title", &title) != B_NO_ERROR ) title = 0;
			mTermTitle = title;
			UpdateTitle();
		} break;
		case TERM_END_SESSION_MSG: {
			bool remote;
			if( message->FindBool("remoteclosed", &remote) == B_NO_ERROR ) {
				if( remote && mExitOnClose ) {
					BMessage quit(B_QUIT_REQUESTED);
					PostMessage(&quit, this);
				}
			}
		}
		case RESIZE_TO_FIT_MSG: {
			ResizeToTerm();
		} break;
		case LOCK_SIZE_MSG: {
			int32 rows=0, cols=0;
			if( message->FindInt32("rows",&rows) == B_NO_ERROR
				&& message->FindInt32("columns",&cols) == B_NO_ERROR ) {
				if( rows > 0 && cols > 0 && terminal ) {
					terminal->TermSetFixedSize(rows,cols);
				}
			}
		} break;
		case SET_FONT_SIZE_MSG: {
			int32 size = 0;
			if( message->FindInt32("size",&size) == B_NO_ERROR ) {
				if( terminal ) {
					BFont font;
					int32 rows=0, cols=0;
					terminal->TermGetViewSize(&rows,&cols);
					terminal->GetFont(&font);
					font.SetSize(size);
					terminal->SetFont(&font);
					ResizeToTerm(rows,cols);
				}
			}
		} break;
		case SET_CRLF_MSG: {
			if( terminal ) {
				terminal->TermSetMode( terminal->TermMode()
									^ terminal->TERM_MODENEWLINE);
			}
		} break;
		case SET_ENTER_MSG: {
			if( terminal ) {
				terminal->TermSetMode( terminal->TermMode()
									^ terminal->TERM_MODEENTERISLF);
			}
		} break;
		case SET_SWAPBSDEL_MSG: {
			if( terminal ) {
				terminal->TermSetMode( terminal->TermMode()
									^ terminal->TERM_MODESWAPBSDEL);
			}
		} break;
		case SET_WRAP_MSG: {
			if( terminal ) {
				terminal->TermSetMode( terminal->TermMode()
									^ terminal->TERM_MODENOWRAP);
			}
		} break;
		case SET_INVERSE_MSG: {
			if( terminal ) {
				terminal->TermSetMode( terminal->TermMode()
									^ terminal->TERM_MODEINVERSE);
				terminal->TermClean();
			}
		} break;
		case SET_PASTEQUERY_MSG: {
			if( terminal ) {
				terminal->SetPasteVerified(!terminal->PasteVerified());
			}
		} break;
		case SET_RMBPASTE_MSG: {
			if( terminal ) {
				terminal->SetRMBPaste(!terminal->RMBPaste());
			}
		} break;
		case SET_ISCROLL_MSG: {
			if( terminal ) {
				terminal->SetAutoScrollMode(
					(ArpTerminal::AutoScrollType)
					(((int)terminal->AutoScrollMode())
						^ terminal->AUTOSCROLL_INPUT) );
			}
		} break;
		case SET_OSCROLL_MSG: {
			if( terminal ) {
				terminal->SetAutoScrollMode(
					(ArpTerminal::AutoScrollType)
					(((int)terminal->AutoScrollMode())
						^ terminal->AUTOSCROLL_OUTPUT) );
			}
		} break;
		case SET_ENCODING_MSG: {
			int32 value;
			if( message->FindInt32("encoding",&value) == B_NO_ERROR ) {
				if( terminal ) {
					terminal->TermSetEncoding(value);
				}
			}
		} break;
		case SOFT_RESET_MSG: {
			if( terminal ) {
				terminal->Emulator().Reset(false);
				terminal->TermClean();
			}
		} break;
		case HARD_RESET_MSG: {
			if( terminal ) {
				terminal->Emulator().Reset(true);
				terminal->TermClean();
			}
		} break;
		case ABOUT_REMOTE_MSG: {
			BView* view = ChildAt(0);
			while( view ) {
				ArpD(cdb << ADH << "Frame of view " << view->Name()
					<< ": " << view->Frame() << endl);
				view = view->NextSibling();
			}
	
			ArpString msg("");
			if( remote ) {
				const ArpRemoteType* type =
					remote->RemoteType();
				if( type ) {
					msg += "Remote protocol: ";
					msg += type->URL;
					if( type->Synonym ) {
						msg += "\nOfficial protocol: ";
						msg += type->Synonym;
					}
					msg += "\nRemote name: ";
					msg += type->LongName;
					const ArpRemoteAddon* addon =
						mRemotes.RemoteAddonForURL(type->URL);
					if( addon ) {
						msg += "\n\nAddon name: ";
						msg += addon->Name;
						msg += "\nCompany: ";
						msg += addon->Company;
						msg += "\nCopyright: ";
						msg += addon->Copyright;
						msg += "\nBuild Date/Time: ";
						msg += addon->BuildDate;
						msg += " / ";
						msg += addon->BuildTime;
						msg += "\nVersion: ";
						msg += addon->Version;
						msg += "\n\n";
						msg += addon->Description;
					}
				}
			}
			if( msg == ArpLIT("") ) msg += ArpLIT("No information available.");
			BAlert* alert = new BAlert("About Device", msg, "Dismiss");
			if( alert ) {
				alert->Go(0);
			}
		} break;
		case ABOUT_EMULATION_MSG: {
			ArpString msg("");
			if( terminal ) {
				const ArpEmulationType* type =
					terminal->Emulator().EmulationType();
				if( type ) {
					msg += "Emulation name: ";
					msg += type->Name;
					if( type->Synonym ) {
						msg += "\nOfficial name: ";
						msg += type->Synonym;
					}
					msg += "\nLong name: ";
					msg += type->LongName;
					const ArpEmulatorAddon* addon =
						mEmulators.EmulatorAddonForType(type->Name);
					if( addon ) {
						msg += "\n\nAddon name: ";
						msg += addon->Name;
						msg += "\nCompany: ";
						msg += addon->Company;
						msg += "\nCopyright: ";
						msg += addon->Copyright;
						msg += "\nBuild Date/Time: ";
						msg += addon->BuildDate;
						msg += " / ";
						msg += addon->BuildTime;
						msg += "\nVersion: ";
						msg += addon->Version;
						msg += "\n\n";
						msg += addon->Description;
					}
				}
			}
			if( msg == ArpLIT("") ) msg += ArpLIT("No information available.");
			BAlert* alert = new BAlert("About Emulator", msg, "Dismiss");
			if( alert ) {
				alert->Go(0);
			}
		} break;
		case TEST_MY_SCREEN: {
			if( terminal ) {
				ArpTerminal& t = *terminal;
				t.TermSetStyle(t.TERM_STYLEPLAIN);
				PrintToTerm(ArpLIT("\r\nExample ArpTerminal Rendering\r\n"));
				PrintToTerm(ArpLIT("\r\nCharacter styles: "));
				t.TermSetStyle(t.TERM_STYLEPLAIN);
				PrintToTerm(ArpLIT("Plain "));
				t.TermSetStyle(t.TERM_STYLEINVERSE);
				PrintToTerm(ArpLIT("Inverse "));
				t.TermSetStyle(t.TERM_STYLEBOLD);
				PrintToTerm(ArpLIT("Bold "));
				t.TermSetStyle(t.TERM_STYLEITALIC);
				PrintToTerm(ArpLIT("Italic "));
				t.TermSetStyle(t.TERM_STYLEUNDERSCORE);
				PrintToTerm(ArpLIT("Underscore "));
				t.TermSetStyle(t.TERM_STYLEINVERSE|t.TERM_STYLEBOLD
						|t.TERM_STYLEITALIC|t.TERM_STYLEUNDERSCORE);
				PrintToTerm(ArpLIT("Combination"));
				t.TermSetStyle(t.TERM_STYLEPLAIN);
				PrintToTerm(ArpLIT("\r\n\r\nColors:\r\n"));
				int i, j;
//				for( i=0; i<t.TERM_NUMTEXTCOLORS; i++ ) {
//					for( j=0; j<t.TERM_NUMTEXTCOLORS; j++ ) {
				for( i=0; i<9; i++ ) {
					for( j=0; j<9; j++ ) {
						t.TermSetBackColor(i);
						t.TermSetForeColor(j);
						//printf("\033[3%dm\033[4%dmCOLOR ",j,i);
						PrintToTerm(ArpLIT("COLOR "),t.TERM_OUTPARTIAL);
					}
					t.TermSetBackColor(0);
					t.TermSetForeColor(0);
					//printf("\033[0m\r\n");
					PrintToTerm(ArpLIT("\r\n"));
				}
					
			}
		} break;
		default:
			inherited::MessageReceived(message);
	}
}

void TerminalWin::MenusBeginning()
{
	if( terminal ) {
		int32 frows=0, fcols=0;
		terminal->TermGetFixedSize(&frows, &fcols);
		
		int32 rows=0, cols=0;
		terminal->TermGetViewSize(&rows,&cols);
		if( frows >= 0 && fcols >= 0 ) {
			rows = frows;
			cols = fcols;
		}
		
		if( winsizemenu ) {
			BMenuItem* found = 0;
			for( int32 i=0; i<winsizemenu->CountItems(); i++ ) {
				BMenuItem* item = winsizemenu->ItemAt(i);
				if( item ) item->SetMarked(false);
				BMessage* msg = item ? item->Message() : 0;
				if( !found && msg ) {
					int32 mrows=0, mcols=0;
					if( msg->FindInt32("rows", &mrows) == B_OK
							&& msg->FindInt32("columns", &mcols) == B_OK ) {
						if( mrows == frows && mcols == fcols ) found = item;
					}
				}
			}
			if( found ) found->SetMarked(true);
			if( found != lockitem ) {
				BMessage* msg = new BMessage(LOCK_SIZE_MSG);
				if( msg ) {
					char labtext[128];
					sprintf(&labtext[0],"Lock At %ld x %ld",rows,cols);
					lockitem->SetLabel(&labtext[0]);
					msg->AddInt32("rows",rows);
					msg->AddInt32("columns",cols);
					lockitem->SetMessage(msg);
				}
			}
		}
		
		if( fontsizemenu ) {
			BFont font;
			terminal->GetFont(&font);
			for( int32 i=0; i<fontsizemenu->CountItems(); i++ ) {
				BMenuItem* item = fontsizemenu->ItemAt(i);
				if( item ) item->SetMarked(false);
				BMessage* msg = item ? item->Message() : 0;
				if( msg ) {
					int32 size=0;
					if( msg->FindInt32("size", &size) == B_OK ) {
						if( font.Size() == size ) item->SetMarked(true);
					}
				}
			}
		}
		
		if( encodingmenu ) {
			int32 encoding = terminal->TermEncoding();
			for( int32 i=0; i<encodingmenu->CountItems(); i++ ) {
				BMenuItem* item = encodingmenu->ItemAt(i);
				if( item ) item->SetMarked(false);
				BMessage* msg = item ? item->Message() : 0;
				if( msg ) {
					int32 menc=0;
					if( msg->FindInt32("encoding", &menc) == B_OK ) {
						if( menc == encoding ) item->SetMarked(true);
					}
				}
			}
		}
	}
	
	if( terminal && emumenu ) terminal->CheckEmulationMenu(emumenu);

	update_modes();
	CheckRemoteMenu();
}

bool TerminalWin::QuitRequested()
{
	if( mCloseConfirm && remote && remote->IsInUse() ) {
		try {
			ArpD(cdb << ADH << "Asking user to end current session." << endl);
			ArpString msg("This will end the current ");
			msg += remote->RemoteType()->LongName;
			msg += " session.\nClose it?";
			BAlert* alert = new BAlert("Close Session", msg, "Cancel", "Close Session",
										0, B_WIDTH_AS_USUAL, B_WARNING_ALERT);
			if( !alert ) return false;
			alert->SetFeel(B_MODAL_SUBSET_WINDOW_FEEL);
			alert->AddToSubset(this);
			if( alert->Go() != 1 ) return false;
		} catch(...) {
			return false;
		}
	}
	if( mSaveOnExit > 0 && mConfigFile.HasFile() ) SaveConfiguration();
	be_app->PostMessage(B_QUIT_REQUESTED);
	return(true);
}
